@@ -22,6 +22,9 @@
  */
 
 #if defined(_WIN32)
+#ifndef UNICODE
+#define UNICODE
+#endif
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
 #ifdef _MSC_VER
@@ -31,7 +34,14 @@
 #endif
 #define LIBNAMETYPE wchar_t*
 #define LIBNAME2CSTR(ENV,JSTR) newWideCString(ENV,JSTR)
-#define LOAD_LIBRARY(NAME) LoadLibraryW(NAME)
+/* See http://msdn.microsoft.com/en-us/library/ms682586(VS.85).aspx:
+ * "Note that the standard search strategy and the alternate search strategy  
+ * specified by LoadLibraryEx with LOAD_WITH_ALTERED_SEARCH_PATH differ in    
+ * just one way: The standard search begins in the calling application's      
+ * directory, and the alternate search begins in the directory of the         
+ * executable module that LoadLibraryEx is loading."                          
+ */
+#define LOAD_LIBRARY(NAME) LoadLibraryExW(NAME, NULL, LOAD_WITH_ALTERED_SEARCH_PATH)
 #define LOAD_ERROR(BUF,LEN) w32_format_error(BUF, LEN)
 #define FREE_LIBRARY(HANDLE) FreeLibrary(HANDLE)
 #define FIND_ENTRY(HANDLE, NAME) GetProcAddress(HANDLE, NAME)
@@ -39,8 +49,8 @@
 #define SET_LAST_ERROR(CODE) SetLastError(CODE)
 static char*
 w32_format_error(char* buf, int len) {
-  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(),
-                0, buf, len, NULL);
+  FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(),
+                 0, buf, len, NULL);
   return buf;
 }
 #else
@@ -52,8 +62,8 @@ w32_format_error(char* buf, int len) {
 #else
 #define LIBNAME2CSTR(ENV,JSTR) newCString(ENV,JSTR)
 #endif
-#define LOAD_LIBRARY(NAME) dlopen(NAME, RTLD_LAZY)
-#define LOAD_ERROR(BUF,LEN) (sprintf(BUF, "%s", dlerror()), BUF)
+#define LOAD_LIBRARY(NAME) dlopen(NAME, RTLD_LAZY|RTLD_GLOBAL)
+#define LOAD_ERROR(BUF,LEN) (snprintf(BUF, LEN, "%s", dlerror()), BUF)
 #define FREE_LIBRARY(HANDLE) dlclose(HANDLE)
 #define FIND_ENTRY(HANDLE, NAME) dlsym(HANDLE, NAME)
 #define GET_LAST_ERROR() errno
@@ -81,9 +91,9 @@ w32_format_error(char* buf, int len) {
 
 #ifdef HAVE_PROTECTION
 static int _protect;
+#undef PROTECT
 #define PROTECT _protect
 #endif
-#include "protect.h"
 
 #ifdef __cplusplus
 extern "C"
@@ -91,9 +101,6 @@ extern "C"
 
 static jboolean preserve_last_error;
 
-#define ON_ERROR() throwByName(env, EError, "Invalid memory access")
-#define PSTART() PROTECTED_START()
-#define PEND() PROTECTED_END(ON_ERROR())
 #define MEMCPY(D,S,L) do { \
   PSTART(); memcpy(D,S,L); PEND(); \
 } while(0)
@@ -191,6 +198,7 @@ static void *getStructureAddress(JNIEnv *, jobject);
 static ffi_type* getStructureType(JNIEnv *, jobject);
 static void update_last_error(JNIEnv*, int);
 
+/** Invokes System.err.println (for debugging only). */
 static void
 println(JNIEnv* env, const char* msg) {
   jclass cls = (*env)->FindClass(env, "java/lang/System");
@@ -229,10 +237,11 @@ dispatch(JNIEnv *env, jobject self, jint callconv, jobjectArray arr,
   nargs = (*env)->GetArrayLength(env, arr);
 
   if (nargs > MAX_NARGS) {
-    sprintf(msg, "Too many arguments (max %ld)", MAX_NARGS);
+    snprintf(msg, sizeof(msg), "Too many arguments (max %ld)", MAX_NARGS);
     throwByName(env, EUnsupportedOperation, msg);
     return;
   }
+
   c_args = (jvalue*)alloca(nargs * sizeof(jvalue));
   array_elements = (struct _array_elements*)
     alloca(nargs * sizeof(struct _array_elements));
@@ -277,7 +286,7 @@ dispatch(JNIEnv *env, jobject self, jint callconv, jobjectArray arr,
         ffi_values[i] = &c_args[i].i;
       }
       else {
-        sprintf(msg, "Unsupported wchar_t size (%d)", (int)sizeof(wchar_t));
+        snprintf(msg, sizeof(msg), "Unsupported wchar_t size (%d)", (int)sizeof(wchar_t));
         throwByName(env, EUnsupportedOperation, msg);
         goto cleanup;
       }
@@ -312,7 +321,8 @@ dispatch(JNIEnv *env, jobject self, jint callconv, jobjectArray arr,
       ffi_types[i] = getStructureType(env, arg);
       ffi_values[i] = c_args[i].l;
       if (!ffi_types[i]) {
-        sprintf(msg, "Structure type info not initialized at argument %d", i);
+        snprintf(msg, sizeof(msg),
+                 "Structure type info not initialized at argument %d", i);
         throwByName(env, EIllegalState, msg);
         goto cleanup;
       }
@@ -322,22 +332,22 @@ dispatch(JNIEnv *env, jobject self, jint callconv, jobjectArray arr,
       ffi_types[i] = &ffi_type_pointer;
       ffi_values[i] = &c_args[i].l;
       if (c_args[i].l == NULL) {
-          c_args[i].l =
-              getBufferArray(env, arg, &array_elements[array_count].array,
-                             &array_elements[array_count].type,
-                             &array_elements[array_count].elems);
-          if (c_args[i].l == NULL) {
-              throwByName(env, EIllegalArgument,
-                          "Buffer arguments must be direct or have a primitive backing array");
-              goto cleanup;
-          }
-          ++array_count;
+        c_args[i].l =
+          getBufferArray(env, arg, &array_elements[array_count].array,
+                         &array_elements[array_count].type,
+                         &array_elements[array_count].elems);
+        if (c_args[i].l == NULL) {
+          throwByName(env, EIllegalArgument,
+                      "Buffer arguments must be direct or have a primitive backing array");
+          goto cleanup;
+        }
+        ++array_count;
       }
     }
     else if ((array_pt = getArrayComponentType(env, arg)) != 0
              && array_pt != 'L') {
       void *ptr = NULL;
-      
+
       switch(array_pt) {
       case 'Z': ptr = (*env)->GetBooleanArrayElements(env, arg, NULL); break;
       case 'B': ptr = (*env)->GetByteArrayElements(env, arg, NULL); break;
@@ -361,9 +371,10 @@ dispatch(JNIEnv *env, jobject self, jint callconv, jobjectArray arr,
       array_elements[array_count++].elems = ptr;
     }
     else {
-      sprintf(msg, "Unsupported type at parameter %d", i);
-      throwByName(env,EIllegalArgument, msg);
-      goto cleanup;
+      // Anything else, pass directly as a pointer
+      c_args[i].l = (void*)arg;
+      ffi_types[i] = &ffi_type_pointer;
+      ffi_values[i] = &c_args[i].l;
     }
   }
   
@@ -371,24 +382,33 @@ dispatch(JNIEnv *env, jobject self, jint callconv, jobjectArray arr,
   case CALLCONV_C:
     abi = FFI_DEFAULT_ABI;
     break;
-#if defined(_WIN32)
+#ifdef _WIN32
   case CALLCONV_STDCALL:
+#ifdef _WIN64
+    // Ignore requests for stdcall on win64
+    abi = FFI_DEFAULT_ABI;
+#else
     abi = FFI_STDCALL;
+#endif
     break;
 #endif // _WIN32
   default:
-    sprintf(msg, "Unrecognized calling convention: %d", (int)callconv);
+    snprintf(msg, sizeof(msg), 
+            "Unrecognized calling convention: %d", (int)callconv);
     throwByName(env, EIllegalArgument, msg);
     goto cleanup;
   }
+
   status = ffi_prep_cif(&cif, abi, nargs, ffi_return_type, ffi_types);
   switch(status) {
   case FFI_BAD_ABI:
-    sprintf(msg, "Invalid calling convention: %d", (int)callconv); 
+    snprintf(msg, sizeof(msg),
+            "Invalid calling convention: %d", (int)callconv); 
     throwByName(env, EIllegalArgument, msg);
     break;
   case FFI_BAD_TYPEDEF:
-    sprintf(msg, "Invalid structure definition (native typedef error)");
+    snprintf(msg, sizeof(msg),
+             "Invalid structure definition (native typedef error)");
     throwByName(env, EIllegalArgument, msg);
     break;
   case FFI_OK: {
@@ -401,7 +421,8 @@ dispatch(JNIEnv *env, jobject self, jint callconv, jobjectArray arr,
     break;
   }
   default:
-    sprintf(msg, "Native call setup failure: error code %d", status);
+    snprintf(msg, sizeof(msg),
+             "Native call setup failure: error code %d", status);
     throwByName(env, EIllegalArgument, msg);
     break;
   }
@@ -494,6 +515,21 @@ Java_com_sun_jna_Function_invokePointer(JNIEnv *env, jobject self,
 }
 
 
+/*
+ * Class:     Function
+ * Method:    invokeObject
+ * Signature: (I[Ljava/lang/Object;)Ljava/lang/Object;
+ */
+JNIEXPORT jobject JNICALL 
+Java_com_sun_jna_Function_invokeObject(JNIEnv *env, jobject self, 
+                                       jint callconv, jobjectArray arr)
+{
+    jvalue result;
+    dispatch(env, self, callconv, arr, &ffi_type_pointer, &result);
+    return result.l;
+}
+
+
 /*
  * Class:     Function
  * Method:    invokeStructure
@@ -505,12 +541,12 @@ Java_com_sun_jna_Function_invokeStructure(JNIEnv *env, jobject self,
                                           jobject result)
 {
   void* memory = getStructureAddress(env, result);
-  ffi_type* type = getStructureType(env, result);
-  if (!type) {
+  ffi_type* rtype = getStructureType(env, result);
+  if (!rtype) {
     throwByName(env, EIllegalState, "Return structure type info not initialized");
   }
   else {
-    dispatch(env, self, callconv, arr, type, memory);
+    dispatch(env, self, callconv, arr, rtype, memory);
   }
   return result;
 }
@@ -622,14 +658,20 @@ Java_com_sun_jna_NativeLibrary_open(JNIEnv *env, jclass cls, jstring lib){
     void *handle = NULL;
     LIBNAMETYPE libname = NULL;
 
-    if ((libname = LIBNAME2CSTR(env, lib)) != NULL) {
-      handle = (void *)LOAD_LIBRARY(libname);
-      if (!handle) {
-        char buf[1024];
-        throwByName(env, EUnsatisfiedLink, LOAD_ERROR(buf, sizeof(buf)));
+    /* dlopen on Unix allows NULL to mean "current process" */
+    if (lib != NULL) {
+      if ((libname = LIBNAME2CSTR(env, lib)) == NULL) {
+        return (jlong)A2L(NULL);
       }
-      free(libname);
     }
+
+    handle = (void *)LOAD_LIBRARY(libname);
+    if (!handle) {
+      char buf[1024];
+      throwByName(env, EUnsatisfiedLink, LOAD_ERROR(buf, sizeof(buf)));
+    }
+    if (libname != NULL)
+      free(libname);
     return (jlong)A2L(handle);
 }
 
@@ -997,7 +1039,7 @@ JNIEXPORT void JNICALL Java_com_sun_jna_Pointer__1read__J_3SII
 JNIEXPORT jbyte JNICALL Java_com_sun_jna_Pointer__1getByte
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    jbyte res;
+    jbyte res = 0;
     MEMCPY(&res, L2A(addr), sizeof(res));
     return res;
 }
@@ -1010,7 +1052,7 @@ JNIEXPORT jbyte JNICALL Java_com_sun_jna_Pointer__1getByte
 JNIEXPORT jchar JNICALL Java_com_sun_jna_Pointer__1getChar
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    wchar_t res;
+    wchar_t res = 0;
     MEMCPY(&res, L2A(addr), sizeof(res));
     return (jchar)res;
 }
@@ -1023,7 +1065,7 @@ JNIEXPORT jchar JNICALL Java_com_sun_jna_Pointer__1getChar
 JNIEXPORT jobject JNICALL Java_com_sun_jna_Pointer__1getPointer
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    void *ptr;
+    void *ptr = NULL;
     MEMCPY(&ptr, L2A(addr), sizeof(ptr));
     return newJavaPointer(env, ptr);
 }
@@ -1047,7 +1089,7 @@ JNIEXPORT jobject JNICALL Java_com_sun_jna_Pointer__1getDirectByteBuffer
 JNIEXPORT jdouble JNICALL Java_com_sun_jna_Pointer__1getDouble
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    jdouble res;
+    jdouble res = 0;
     MEMCPY(&res, L2A(addr), sizeof(res));
     return res;
 }
@@ -1060,7 +1102,7 @@ JNIEXPORT jdouble JNICALL Java_com_sun_jna_Pointer__1getDouble
 JNIEXPORT jfloat JNICALL Java_com_sun_jna_Pointer__1getFloat
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    jfloat res;
+    jfloat res = 0;
     MEMCPY(&res, L2A(addr), sizeof(res));
     return res;
 }
@@ -1073,7 +1115,7 @@ JNIEXPORT jfloat JNICALL Java_com_sun_jna_Pointer__1getFloat
 JNIEXPORT jint JNICALL Java_com_sun_jna_Pointer__1getInt
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    jint res;
+    jint res = 0;
     MEMCPY(&res, L2A(addr), sizeof(res));
     return res;
 }
@@ -1086,7 +1128,7 @@ JNIEXPORT jint JNICALL Java_com_sun_jna_Pointer__1getInt
 JNIEXPORT jlong JNICALL Java_com_sun_jna_Pointer__1getLong
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    jlong res;
+    jlong res = 0;
     MEMCPY(&res, L2A(addr), sizeof(res));
     return res;
 }
@@ -1099,7 +1141,7 @@ JNIEXPORT jlong JNICALL Java_com_sun_jna_Pointer__1getLong
 JNIEXPORT jshort JNICALL Java_com_sun_jna_Pointer__1getShort
     (JNIEnv *env, jclass cls, jlong addr)
 {
-    jshort res;
+    jshort res = 0;
     MEMCPY(&res, L2A(addr), sizeof(res));
     return res;
 }
@@ -1379,7 +1421,7 @@ newJavaString(JNIEnv *env, const char *ptr, jboolean wide)
 
     if (wide) {
         // TODO: proper conversion from native wchar_t to jchar
-        int len = wcslen((const wchar_t*)ptr);
+        int len = (int)wcslen((const wchar_t*)ptr);
         if (sizeof(jchar) != sizeof(wchar_t)) {
             jchar* buf = (jchar*)alloca(len * sizeof(jchar));
             int i;
@@ -1394,7 +1436,7 @@ newJavaString(JNIEnv *env, const char *ptr, jboolean wide)
     }
     else {
         jbyteArray bytes = 0;
-        int len = strlen(ptr);
+        int len = (int)strlen(ptr);
 
         bytes = (*env)->NewByteArray(env, len);
         if (bytes != 0) {
@@ -1534,39 +1576,32 @@ do { jboolean cpy; \
     *elemp = ptr;
     ptr = (char *)ptr + offset;
   }
+
   return ptr;
 }
 
 
 /*
  * Class:     Native
- * Method:    pointerSize
- * Signature: ()I
+ * Method:    sizeof
+ * Signature: (I)I
  */
 JNIEXPORT jint JNICALL 
-Java_com_sun_jna_Native_pointerSize(JNIEnv *env, jclass cls)
+Java_com_sun_jna_Native_sizeof(JNIEnv *env, jclass cls, jint type)
 {
-  return sizeof(void *);
-}
-
-/*
- * Class:     Native
- * Method:    longSize
- * Signature: ()I
- */
-JNIEXPORT jint JNICALL 
-Java_com_sun_jna_Native_longSize(JNIEnv *env, jclass cls) {
-  return sizeof(long);
-}
-
-/*
- * Class:     Native
- * Method:    wideCharSize
- * Signature: ()I
- */
-JNIEXPORT jint JNICALL 
-Java_com_sun_jna_Native_wideCharSize(JNIEnv *env, jclass cls) {
-  return sizeof(wchar_t);
+  switch(type) {
+  case com_sun_jna_Native_TYPE_VOIDP: return sizeof(void*);
+  case com_sun_jna_Native_TYPE_LONG: return sizeof(long);
+  case com_sun_jna_Native_TYPE_WCHAR_T: return sizeof(wchar_t);
+  case com_sun_jna_Native_TYPE_SIZE_T: return sizeof(size_t);
+  default:
+    {
+      char msg[1024];
+      snprintf(msg, sizeof(msg), "Invalid sizeof type %d", (int)type);
+      throwByName(env, EIllegalArgument, msg);
+      return -1;
+    }
+  }
 }
 
 /** Initialize com.sun.jna classes separately from the library load to
@@ -1627,7 +1662,7 @@ Java_com_sun_jna_Native_initIDs(JNIEnv *env, jclass cls) {
   }
 
   // Initialize type fields within Structure.FFIType
-  {
+  else {
 #define CFFITYPE "com/sun/jna/Structure$FFIType$FFITypes"
     jclass cls = (*env)->FindClass(env, CFFITYPE);
     jfieldID fid;
@@ -1652,7 +1687,7 @@ Java_com_sun_jna_Native_initIDs(JNIEnv *env, jclass cls) {
       return;
     }
     for (i=0;i < sizeof(fields)/sizeof(fields[0]);i++) {
-      sprintf(field, "ffi_type_%s", fields[i]);
+      snprintf(field, sizeof(field), "ffi_type_%s", fields[i]);
       fid = (*env)->GetStaticFieldID(env, cls, field, "Lcom/sun/jna/Pointer;");
       if (!fid) {
         throwByName(env, EUnsatisfiedLink, field);
@@ -1667,13 +1702,13 @@ Java_com_sun_jna_Native_initIDs(JNIEnv *env, jclass cls) {
 #define JAWT_HEADLESS_HACK
 #ifdef _WIN32
 #define JAWT_NAME "jawt.dll"
-#define METHOD_NAME (sizeof(void*)==4?"_JAWT_GetAWT@8":"_JAWT_GetAWT@16")
+#define METHOD_NAME (sizeof(void*)==4?"_JAWT_GetAWT@8":"JAWT_GetAWT")
 #else
 #define JAWT_NAME "libjawt.so"
 #define METHOD_NAME "JAWT_GetAWT"
 #endif
 static void* jawt_handle = NULL;
-static jboolean JNICALL (*pJAWT_GetAWT)(JNIEnv*,JAWT*);
+static jboolean (JNICALL *pJAWT_GetAWT)(JNIEnv*,JAWT*);
 #define JAWT_GetAWT (*pJAWT_GetAWT)
 #endif
 
@@ -1714,16 +1749,20 @@ Java_com_sun_jna_Native_getWindowHandle0(JNIEnv *env, jclass classp, jobject w)
         if (java_home != NULL) {
           if ((prop = newWideCString(env, java_home)) != NULL) {
             const wchar_t* suffix = L"/bin/jawt.dll";
-            path = (wchar_t*)alloca((wcslen(prop) + wcslen(suffix) + 1)
-                                    * sizeof(wchar_t));
+            size_t len = wcslen(prop) + wcslen(suffix) + 1;
+            path = (wchar_t*)alloca(len * sizeof(wchar_t));
+#ifdef _MSC_VER
+            swprintf(path, len, L"%s%s", prop, suffix);
+#else
             swprintf(path, L"%s%s", prop, suffix);
+#endif
             free(prop);
           }
         }
       }
     }
-#undef LOAD_LIBRARY
-#define LOAD_LIBRARY(X) LoadLibraryW(path)
+#undef JAWT_NAME
+#define JAWT_NAME path
 #endif
     if ((jawt_handle = LOAD_LIBRARY(JAWT_NAME)) == NULL) {
       char msg[1024];
@@ -1732,7 +1771,7 @@ Java_com_sun_jna_Native_getWindowHandle0(JNIEnv *env, jclass classp, jobject w)
     }
     if ((pJAWT_GetAWT = (void*)FIND_ENTRY(jawt_handle, METHOD_NAME)) == NULL) {
       char msg[1024], buf[1024];
-      sprintf(msg, "Error looking up %s: %s",
+      snprintf(msg, sizeof(msg), "Error looking up %s: %s",
               METHOD_NAME, LOAD_ERROR(buf, sizeof(buf)));
       throwByName(env, EUnsatisfiedLink, msg);
       return -1;
@@ -1768,7 +1807,7 @@ Java_com_sun_jna_Native_getWindowHandle0(JNIEnv *env, jclass classp, jobject w)
         // FIXME this kills the VM if the window is not realized;
         // if not, wdsi might be a bogus, non-null value
         // TODO: fix JVM (right) or ensure window is realized (done in Java)
-        handle = (jint)wdsi->hwnd;
+        handle = A2L(wdsi->hwnd);
         if (!handle) {
           throwByName(env, EIllegalState, "Can't get HWND");
         }
@@ -1828,14 +1867,19 @@ Java_com_sun_jna_Native_setProtected(JNIEnv *env, jclass classp, jboolean protec
 #endif
 }
 
-JNIEXPORT jboolean JNICALL
-Java_com_sun_jna_Native_isProtected(JNIEnv *env, jclass classp) {
-#ifdef HAVE_PROTECTION
+jboolean
+is_protected() {
+#ifdef HAVE_PROTECTION  
   if (_protect) return JNI_TRUE;
 #endif
   return JNI_FALSE;
 }
 
+JNIEXPORT jboolean JNICALL
+Java_com_sun_jna_Native_isProtected(JNIEnv *env, jclass classp) {
+  return is_protected();
+}
+
 JNIEXPORT void JNICALL
 Java_com_sun_jna_Native_setPreserveLastError(JNIEnv *env, jclass classp, jboolean preserve) {
   preserve_last_error = preserve;
@@ -2058,9 +2102,8 @@ get_ffi_type(JNIEnv* env, jclass cls, char jtype) {
     return getStructureType(env, s);
   }
   case '*':
-    return &ffi_type_pointer;
   default:
-    return NULL;
+    return &ffi_type_pointer;
   }
 }
 
@@ -2082,7 +2125,136 @@ get_ffi_rtype(JNIEnv* env, jclass cls, char jtype) {
     return get_ffi_type(env, cls, jtype);
   }
 }
-  
+
+typedef struct _method_data {
+  ffi_cif cif;
+  ffi_cif closure_cif;
+  void*   fptr;
+  ffi_type** arg_types;
+} method_data;
+
+// VM vectors to this callback, which calls native code
+static void
+method_handler(ffi_cif* cif, void* resp, void** argp, void *cdata) {
+  JNIEnv* env = (JNIEnv*)*(void **)argp[0];
+  method_data *data = (method_data*)cdata;
+  // ignore first two arguments, which are pointers
+  void** args = argp + 2;
+  unsigned i;
+
+  // TODO: conversions here otherwise done in JNA interface code
+  // TODO: Pointer, structure by value, primitive arrays
+  for (i=0;i < data->cif.nargs;i++) {
+    if (data->cif.arg_types[i] == &ffi_type_pointer) {
+      if (*(void **)args[i]) {
+        *(void **)args[i] = getNativeAddress(env, *(void **)args[i]);
+      }
+    }
+  }
+
+  ffi_call(&data->cif, FFI_FN(data->fptr), resp, args);
+
+  if (data->cif.rtype == &ffi_type_pointer) {
+    *(void **)resp = newJavaPointer(env, *(void **)resp);
+  }
+}
+
+JNIEXPORT void JNICALL
+Java_com_sun_jna_Native_unregister(JNIEnv *env, jclass ncls, jclass cls, jlongArray handles) {
+  jlong* data = (*env)->GetLongArrayElements(env, handles, NULL);
+  int count = (*env)->GetArrayLength(env, handles);
+  while (count-- > 0) {
+    method_data* md = (method_data*)L2A(data[count]);
+    free(md->arg_types);
+    free(md);
+  }
+  (*env)->ReleaseLongArrayElements(env, handles, data, 0);
+  (*env)->UnregisterNatives(env, cls);
+}
+
+JNIEXPORT jlong JNICALL
+Java_com_sun_jna_Native_registerMethod(JNIEnv *env, jclass ncls,
+                                       jclass cls, jstring name,
+                                       jstring signature,
+                                       jobject function, jint cc)
+{
+  int argc = 0;
+  char* cname = newCStringUTF8(env, name);
+  char* sig = newCStringUTF8(env, signature);
+  void *code;
+  void *closure;
+  method_data* data = malloc(sizeof(method_data));
+  ffi_cif* closure_cif = &data->closure_cif;
+  int status;
+  int i;
+  int abi = FFI_DEFAULT_ABI; 
+  char* tmp;
+  ffi_type* rtype;
+  ffi_type** arg_types = alloca(sizeof(ffi_type*)*MAX_NARGS);
+#if defined(_WIN32) && !defined(_WIN64)
+  if (cc == CALLCONV_STDCALL) abi = FFI_STDCALL;
+#endif
+
+  // Parse the signature for types
+  for (tmp=sig;*tmp;tmp++) {
+    if (*tmp == '(' || *tmp == '[') continue;
+    else if (*tmp == ')') {
+      rtype = get_ffi_rtype(env, NULL, *(tmp+1));
+      break;
+    }
+    else if (*tmp == 'L') {
+      arg_types[argc++] = &ffi_type_pointer;
+      while (*tmp != ';') ++tmp;
+    }
+    else {
+      arg_types[argc++] = get_ffi_type(env, NULL, *tmp);
+    }
+  }
+
+  data->arg_types = malloc(sizeof(ffi_type*)*(argc+2));
+  data->arg_types[0] = &ffi_type_pointer;
+  data->arg_types[1] = &ffi_type_pointer;
+  for (i=0;i < argc;i++) {
+    data->arg_types[i+2] = arg_types[i];
+  }
+  data->fptr = getNativeAddress(env, function);
+
+  status = ffi_prep_cif(closure_cif, abi, argc+2, rtype, data->arg_types);  
+  if (status != FFI_OK) {
+    throwByName(env, EError, "Native method mapping failed");
+    goto cleanup;
+  }
+
+  status = ffi_prep_cif(&data->cif, abi, argc, rtype, &data->arg_types[2]);
+  if (status != FFI_OK) {
+    throwByName(env, EError, "Native method setup failed");
+    goto cleanup;
+  }
+
+  closure = ffi_closure_alloc(sizeof(ffi_closure), &code);
+  status = ffi_prep_closure_loc(closure, closure_cif, method_handler, data, code);
+  if (status != FFI_OK) {
+    throwByName(env, EError, "Native method linkage failed");
+    goto cleanup;
+  }
+
+  {
+    JNINativeMethod m  = { cname, sig, code };
+    (*env)->RegisterNatives(env, cls, &m, 1);
+  }
+
+ cleanup:
+  if (status != FFI_OK) {
+    free(data->arg_types);
+    free(data);
+    data = NULL;
+  }
+  free(cname);
+  free(sig);
+
+  return A2L(data);
+}
+
 #ifdef __cplusplus
 }
 #endif
