@@ -1,4 +1,4 @@
-/* Copyright (c) 2007 Timothy Wall, All Rights Reserved
+/* Copyright (c) 2007-2008 Timothy Wall, All Rights Reserved
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,9 +17,12 @@ import java.lang.reflect.InvocationHandler;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
 import java.lang.reflect.Proxy;
+import java.util.Arrays;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.Iterator;
 import java.util.Map;
+import java.util.Set;
 import java.util.WeakHashMap;
 
 /** Provides a reference to an association between a native callback closure
@@ -28,12 +31,21 @@ import java.util.WeakHashMap;
 
 class CallbackReference extends WeakReference {
     
-    private interface NativeFunctionProxy { }
-    
     static final Map callbackMap = new WeakHashMap();
     static final Map altCallbackMap = new WeakHashMap();
-    private static final Map nativeStrings = new WeakHashMap();
+    static final Map allocations = new WeakHashMap();
+    private static final Method PROXY_CALLBACK_METHOD;
     
+    static {
+        try {
+            PROXY_CALLBACK_METHOD = CallbackProxy.class.getMethod("callback", new Class[] { Object[].class });
+        }
+        catch(Exception e) {
+            throw new Error("Error looking up CallbackProxy.callback() method");
+        }
+    }
+
+
     /** Return a Callback associated with the given function pointer.
      * If the pointer refers to a Java callback trampoline, return the original
      * Java Callback.  Otherwise, return a proxy to the native function pointer.
@@ -60,7 +72,7 @@ class CallbackReference extends WeakReference {
                     ? Function.ALT_CONVENTION : Function.C_CONVENTION;
                 Map options = Native.getLibraryOptions(type);
                 NativeFunctionHandler h = new NativeFunctionHandler(p, ctype, options);
-                Callback cb = (Callback)Proxy.newProxyInstance(type.getClassLoader(), new Class[] { type, NativeFunctionProxy.class }, h);
+                Callback cb = (Callback)Proxy.newProxyInstance(type.getClassLoader(), new Class[] { type }, h);
                 h.options.put(Function.OPTION_INVOKING_METHOD, getCallbackMethod(cb));
                 map.put(cb, null);
                 return cb;
@@ -69,30 +81,17 @@ class CallbackReference extends WeakReference {
         return null;
     }
     
-    private static Class getCallbackClass(Class type) {
-        Class[] ifaces = type.getInterfaces();
-        for (int i=0;i < ifaces.length;i++) {
-            if (Callback.class.isAssignableFrom(ifaces[i])) {
-                type = ifaces[i];
-                break;
-            }
-        }
-        return type;
-    }
-    
     Pointer cbstruct;
     // Keep a reference to avoid premature GC
     CallbackProxy proxy;
     private CallbackReference(Callback callback, int callingConvention) {
         super(callback);
-        Class type = getCallbackClass(callback.getClass());
-        TypeMapper mapper = Native.getTypeMapper(type);
-        Method m = getCallbackMethod(callback);
+        TypeMapper mapper = Native.getTypeMapper(Native.findCallbackClass(callback.getClass()));
         if (callback instanceof CallbackProxy) {
             proxy = (CallbackProxy)callback;
         }
         else {
-            proxy = new DefaultCallbackProxy(m, mapper);
+            proxy = new DefaultCallbackProxy(getCallbackMethod(callback), mapper);
         }
 
         // Generate a list of parameter types that the native code can 
@@ -126,9 +125,7 @@ class CallbackReference extends WeakReference {
                 + " requires custom type conversion";
             throw new IllegalArgumentException(msg);
         }
-
-        Method proxyMethod = getCallbackMethod(proxy);
-        cbstruct = createNativeCallback(proxy, proxyMethod,  
+        cbstruct = createNativeCallback(proxy, PROXY_CALLBACK_METHOD,  
                                         nativeParamTypes, returnType,
                                         callingConvention);
     }
@@ -145,27 +142,50 @@ class CallbackReference extends WeakReference {
         }
         else if (cls == String.class 
                  || cls == WString.class
+                 || cls == String[].class
+                 || cls == WString[].class
                  || Callback.class.isAssignableFrom(cls)) {
             return Pointer.class;
         }
         return cls;
     }
     
+    private static Method checkMethod(Method m) {
+        if (m.getParameterTypes().length > Function.MAX_NARGS) {
+            String msg = "Method signature exceeds the maximum "
+                + "parameter count: " + m;
+            throw new IllegalArgumentException(msg);
+        }
+        return m;
+    }
+
     private static Method getCallbackMethod(Callback callback) {
-        Method[] mlist = callback.getClass().getMethods();
-        for (int mi=0;mi < mlist.length;mi++) {
-            Method m = mlist[mi];
-            if (Callback.METHOD_NAME.equals(m.getName())) {
-                if (m.getParameterTypes().length > Function.MAX_NARGS) {
-                    String msg = "Method signature exceeds the maximum "
-                        + "parameter count: " + m;
-                    throw new IllegalArgumentException(msg);
-                }
-                return m;
+        // Look at only public methods defined by the Callback class
+        Class cls = Native.findCallbackClass(callback.getClass());
+        Method[] pubMethods = cls.getDeclaredMethods();
+        Method[] classMethods = cls.getMethods();
+        Set pmethods = new HashSet(Arrays.asList(pubMethods));
+        pmethods.retainAll(Arrays.asList(classMethods));
+
+        // Remove Object methods disallowed as callback method names 
+        for (Iterator i=pmethods.iterator();i.hasNext();) {
+            Method m = (Method)i.next();
+            if (Callback.FORBIDDEN_NAMES.contains(m.getName())) {
+                i.remove();
             }
         }
-        String msg = "Callback must implement method named '"
-            + Callback.METHOD_NAME + "'";
+        Method[] methods = (Method[])pmethods.toArray(new Method[pmethods.size()]);
+        if (methods.length == 1) {
+            return checkMethod(methods[0]);
+        }
+        for (int i=0;i < methods.length;i++) {
+            Method m = methods[i];
+            if (Callback.METHOD_NAME.equals(m.getName())) {
+                return checkMethod(m);
+            }
+        }
+        String msg = "Callback must implement a single public method, "
+            + "or one public method named '" + Callback.METHOD_NAME + "'";
         throw new IllegalArgumentException(msg);
     }
     
@@ -185,10 +205,11 @@ class CallbackReference extends WeakReference {
     }
 
     private static Pointer getNativeFunctionPointer(Callback cb) {
-        if (cb instanceof NativeFunctionProxy) {
-            NativeFunctionHandler handler = 
-                (NativeFunctionHandler)Proxy.getInvocationHandler(cb);
-            return handler.getPointer();
+        if (Proxy.isProxyClass(cb.getClass())) {
+            Object handler = Proxy.getInvocationHandler(cb);
+            if (handler instanceof NativeFunctionHandler) {
+                return ((NativeFunctionHandler)handler).getPointer();
+            }
         }
         return null;
     }
@@ -249,7 +270,7 @@ class CallbackReference extends WeakReference {
             }
         }
         
-        private Object callback_inner(Object[] args) {
+        private Object invokeCallback(Object[] args) {
             Class[] paramTypes = callbackMethod.getParameterTypes();
             Object[] callbackArgs = new Object[args.length];
             
@@ -272,13 +293,13 @@ class CallbackReference extends WeakReference {
                     result = convertResult(callbackMethod.invoke(cb, callbackArgs));
                 }
                 catch (IllegalArgumentException e) {
-                    e.printStackTrace();
+                    Native.getCallbackExceptionHandler().uncaughtException(cb, e);
                 }
                 catch (IllegalAccessException e) {
-                    e.printStackTrace();
+                    Native.getCallbackExceptionHandler().uncaughtException(cb, e);
                 }
                 catch (InvocationTargetException e) {
-                    e.printStackTrace();
+                    Native.getCallbackExceptionHandler().uncaughtException(cb, e.getTargetException());
                 }
             }
             return result;
@@ -291,10 +312,10 @@ class CallbackReference extends WeakReference {
          */
         public Object callback(Object[] args) {
             try {
-                return callback_inner(args);
+                return invokeCallback(args);
             }
             catch (Throwable t) {
-                t.printStackTrace();
+                Native.getCallbackExceptionHandler().uncaughtException(getCallback(), t);
                 return null;
             }
         }
@@ -310,19 +331,26 @@ class CallbackReference extends WeakReference {
                 else if (dstType == WString.class) {
                     value = new WString(((Pointer)value).getString(0, true));
                 }
+                else if (dstType == String[].class
+                         || dstType == WString[].class) {
+                    value = ((Pointer)value).getStringArray(0, dstType == WString[].class);
+                }
                 else if (Callback.class.isAssignableFrom(dstType)) {
                     value = getCallback(dstType, (Pointer)value);
                 }
                 else if (Structure.class.isAssignableFrom(dstType)) {
                     Structure s = Structure.newInstance(dstType);
-                    Pointer old = s.getPointer();
-                    s.useMemory((Pointer)value);
-                    s.read();
-                    // If by value, don't hold onto the pointer
+                    // If passed by value, don't hold onto the pointer, which
+                    // is only valid for the duration of the callback call
                     if (Structure.ByValue.class.isAssignableFrom(dstType)) {
-                        s.useMemory(old);
-                        s.write();
+                        byte[] buf = new byte[s.size()];
+                        ((Pointer)value).read(0, buf, 0, buf.length);
+                        s.getPointer().write(0, buf, 0, buf.length);
                     }
+                    else {
+                        s.useMemory((Pointer)value);
+                    }
+                    s.read();
                     value = s;
                 }
             }
@@ -347,15 +375,23 @@ class CallbackReference extends WeakReference {
                 return ((Structure)value).getPointer();
             }
             else if (cls == boolean.class || cls == Boolean.class) {
-                return new Integer(Boolean.TRUE.equals(value)?-1:0);
+                return Boolean.TRUE.equals(value) ? 
+                    Function.INTEGER_TRUE : Function.INTEGER_FALSE;
             }
             else if (cls == String.class || cls == WString.class) {
-                // Store in a weak hash map to delay GC until string
-                // itself is GC'd.
                 NativeString ns = new NativeString(value.toString(), cls == WString.class);
-                nativeStrings.put(value, ns);
+                // Delay GC until string itself is GC'd.
+                allocations.put(value, ns);
                 return ns.getPointer();
             }
+            else if (cls == String[].class || cls == WString.class) {
+                StringArray sa = cls == String[].class
+                    ? new StringArray((String[])value)
+                    : new StringArray((WString[])value);
+                // Delay GC until array itself is GC'd.
+                allocations.put(value, sa);
+                return sa;
+            }
             else if (Callback.class.isAssignableFrom(cls)) {
                 return getFunctionPointer((Callback)value);
             }
@@ -383,7 +419,7 @@ class CallbackReference extends WeakReference {
                     String str = super.getName();
                     if (options.containsKey(Function.OPTION_INVOKING_METHOD)) {
                         Method m = (Method)options.get(Function.OPTION_INVOKING_METHOD);
-                        Class cls = getCallbackClass(m.getDeclaringClass());
+                        Class cls = Native.findCallbackClass(m.getDeclaringClass());
                         str += " (" + cls.getName() + ")";
                     }
                     return str;
