@@ -13,7 +13,6 @@ package com.sun.jna;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
 import java.util.Collections;
-import java.util.HashMap;
 import java.util.Map;
 
 /**
@@ -46,6 +45,9 @@ public class Function extends Pointer {
     /** First alternate convention (currently used only for w32 stdcall). */
     public static final int ALT_CONVENTION = 1;
 
+    static final Integer INTEGER_TRUE = new Integer(-1);
+    static final Integer INTEGER_FALSE = new Integer(0);
+
     /** 
      * Obtain a <code>Function</code> representing a native 
      * function that follows the standard "C" calling convention.
@@ -89,8 +91,8 @@ public class Function extends Pointer {
     // Keep a reference to the NativeLibrary so it does not get garbage collected
     // until the function is
     private NativeLibrary library;
-    private String functionName;
-    private int callingConvention;
+    private final String functionName;
+    final int callingConvention;
 
     /** For internal JNA use. */
     static final String OPTION_INVOKING_METHOD = "invoking-method";
@@ -199,8 +201,10 @@ public class Function extends Pointer {
         TypeMapper mapper = 
             (TypeMapper)options.get(Library.OPTION_TYPE_MAPPER);
         Method invokingMethod = (Method)options.get(OPTION_INVOKING_METHOD);
+        boolean allowObjects = Boolean.TRUE.equals(options.get(Library.OPTION_ALLOW_OBJECTS));
         for (int i=0; i < args.length; i++) {
-            args[i] = convertArgument(args, i, invokingMethod, mapper);
+            args[i] = convertArgument(args, i, invokingMethod,
+                                      mapper, allowObjects);
         }
         
         Class nativeType = returnType;
@@ -217,7 +221,7 @@ public class Function extends Pointer {
             }
         }
 
-        Object result = invoke(args, nativeType);
+        Object result = invoke(args, nativeType, allowObjects);
 
         // Convert the result to a custom value/type if appropriate
         if (resultConverter != null) {
@@ -239,7 +243,9 @@ public class Function extends Pointer {
                     continue;
                 if (inArg instanceof Structure) {
                     if (!(inArg instanceof Structure.ByValue)) {
-                        ((Structure)inArg).read();
+                        if (((Structure)inArg).getAutoRead()) {
+                            ((Structure)inArg).read();
+                        }
                     }
                 }
                 else if (args[i] instanceof PostCallRead) {
@@ -259,7 +265,9 @@ public class Function extends Pointer {
                 else if (Structure[].class.isAssignableFrom(inArg.getClass())) {
                     Structure[] ss = (Structure[])inArg;
                     for (int si=0;si < ss.length;si++) {
-                        ss[si].read();
+                        if (ss[si].getAutoRead()) {
+                            ss[si].read();
+                        }
                     }
                 }
             }
@@ -268,8 +276,8 @@ public class Function extends Pointer {
         return result;
     }
 
-    /** @see NativeLibrary#NativeLibrary(String,String,long) implementation */
-    Object invoke(Object[] args, Class returnType) {
+    /** @see NativeLibrary#NativeLibrary(String,String,long,Map) implementation */
+    Object invoke(Object[] args, Class returnType, boolean allowObjects) {
         Object result = null;
         if (returnType == null || returnType==void.class || returnType==Void.class) {
             invokeVoid(callingConvention, args);
@@ -304,7 +312,9 @@ public class Function extends Pointer {
         }
         else if (returnType==WString.class) {
             String s = invokeString(callingConvention, args, true);
-            result = s != null ? new WString(s) : null;
+            if (s != null) {
+                result = new WString(s);
+            }
         }
         else if (Pointer.class.isAssignableFrom(returnType)) {
             result = invokePointer(callingConvention, args);
@@ -314,7 +324,9 @@ public class Function extends Pointer {
                 Structure s = 
                     invokeStructure(callingConvention, args, 
                                     Structure.newInstance(returnType));
-                s.read();
+                if (s.getAutoRead()) {
+                    s.read();
+                }
                 result = s;
             }
             else {
@@ -322,7 +334,9 @@ public class Function extends Pointer {
                 if (result != null) {
                     Structure s = Structure.newInstance(returnType);
                     s.useMemory((Pointer)result);
-                    s.read();
+                    if (s.getAutoRead()) {
+                        s.read();
+                    }
                     result = s;
                 }
             }
@@ -333,6 +347,38 @@ public class Function extends Pointer {
                 result = CallbackReference.getCallback(returnType, (Pointer)result);
             }
         }
+        else if (returnType==String[].class) {
+            Pointer p = invokePointer(callingConvention, args);
+            if (p != null) {
+                result = p.getStringArray(0);
+            }
+        }
+        else if (returnType==WString[].class) {
+            Pointer p = invokePointer(callingConvention, args);
+            if (p != null) {
+                String[] arr = p.getStringArray(0, true);
+                WString[] warr = new WString[arr.length];
+                for (int i=0;i < arr.length;i++) {
+                    warr[i] = new WString(arr[i]);
+                }
+                result = warr;
+            }
+        }
+        else if (returnType==Pointer[].class) {
+            Pointer p = invokePointer(callingConvention, args);
+            if (p != null) {
+                result = p.getPointerArray(0);
+            }
+        }
+        else if (allowObjects) {
+            result = invokeObject(callingConvention, args);
+            if (result != null
+                && !returnType.isAssignableFrom(result.getClass())) {
+                throw new IllegalArgumentException("Return type " + returnType
+                                                   + " does not match result "
+                                                   + result.getClass());
+            }
+        }
         else {
             throw new IllegalArgumentException("Unsupported return type "
                                                + returnType
@@ -341,7 +387,9 @@ public class Function extends Pointer {
         return result;
     }
     
-    private Object convertArgument(Object[] args, int index, Method invokingMethod, TypeMapper mapper) { 
+    private Object convertArgument(Object[] args, int index,
+                                   Method invokingMethod, TypeMapper mapper,
+                                   boolean allowObjects) { 
         Object arg = args[index];
         if (arg != null) {
             Class type = arg.getClass();
@@ -370,7 +418,9 @@ public class Function extends Pointer {
         // Convert Structures to native pointers 
         if (arg instanceof Structure) {
             Structure struct = (Structure)arg;
-            struct.write();
+            if (struct.getAutoWrite()) {
+                struct.write();
+            }
             if (struct instanceof Structure.ByValue) {
             	// Double-check against the method signature, if available
                 Class ptype = struct.getClass();
@@ -415,7 +465,7 @@ public class Function extends Pointer {
         // Default conversion of boolean to int; if you want something
         // different, use a ToNativeConverter
         else if (arg instanceof Boolean) {
-            return new Integer(Boolean.TRUE.equals(arg) ? -1 : 0);
+            return Boolean.TRUE.equals(arg) ? INTEGER_TRUE : INTEGER_FALSE;
         }
         else if (String[].class == argClass) {
             return new StringArray((String[])arg);
@@ -456,23 +506,31 @@ public class Function extends Pointer {
             else {
                 Pointer base = ss[0].getPointer();
                 int size = ss[0].size();
-                ss[0].write();
+                if (ss[0].getAutoWrite()) {
+                    ss[0].write();
+                }
                 for (int si=1;si < ss.length;si++) {
                     if (ss[si].getPointer().peer != base.peer + size*si) {
                         String msg = "Structure array elements must use"
                             + " contiguous memory (at element index " + si + ")";     
                         throw new IllegalArgumentException(msg);
                     }
-                    ss[si].write();
+                    if (ss[si].getAutoWrite()) {
+                        ss[si].write();
+                    }
                 }
                 return base;
             }
         }
         else if (argClass.isArray()){
+            // TODO: handle array of NativeMapped
             throw new IllegalArgumentException("Unsupported array argument type: " 
                                                + argClass.getComponentType());
         }
-        if (arg != null && !Native.isSupportedNativeType(arg.getClass())) {
+        else if (allowObjects) {
+            return arg;
+        }
+        else if (!Native.isSupportedNativeType(arg.getClass())) {
             throw new IllegalArgumentException("Unsupported argument type "
                                                + arg.getClass().getName()
                                                + " at parameter " + index
@@ -579,7 +637,8 @@ public class Function extends Pointer {
     private native Pointer invokePointer(int callingConvention, Object[] args);
     
     /**
-     * Call the native function being represented by this object
+     * Call the native function being represented by this object, returning
+     * a struct by value.
      *
      * @param   callingConvention calling convention to be used
      * @param   args
@@ -590,6 +649,17 @@ public class Function extends Pointer {
     private native Structure invokeStructure(int callingConvention, Object[] args,
                                              Structure result);
 
+    /**
+     * Call the native function being represented by this object, returning
+     * a Java <code>Object</code>.
+     *
+     * @param   callingConvention calling convention to be used
+     * @param   args
+     *          Arguments to pass to the native function
+     * @return  The returned Java <code>Object</code>
+     */
+    private native Object invokeObject(int callingConvention, Object[] args);
+
     /** Provide a human-readable representation of this object. */
     public String toString() {
         if (library != null) {
@@ -599,6 +669,13 @@ public class Function extends Pointer {
         return "native function@0x" + Long.toHexString(peer);
     }
 
+    /** Convenience method for
+     * {@link #invoke(Class,Object[]) invokeObject(Object.class, args)}.
+     */
+    public Object invokeObject(Object[] args) {
+        return invoke(Object.class, args);
+    }
+
     /** Convenience method for 
      * {@link #invoke(Class,Object[]) invoke(Pointer.class, args)}.
      */
@@ -708,7 +785,7 @@ public class Function extends Pointer {
     }
     
     private static class PointerArray extends Memory implements PostCallRead {
-        private Pointer[] original;
+        private final Pointer[] original;
         public PointerArray(Pointer[] arg) {
             super(Pointer.SIZE * (arg.length+1));
             this.original = arg;
