@@ -1,4 +1,4 @@
-/* Copyright (c) 2007 Timothy Wall, All Rights Reserved
+/* Copyright (c) 2007, 2008, 2009 Timothy Wall, All Rights Reserved
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -15,7 +15,6 @@ package com.sun.jna;
 import java.awt.Component;
 import java.awt.GraphicsEnvironment;
 import java.awt.HeadlessException;
-import java.awt.Toolkit;
 import java.awt.Window;
 import java.io.File;
 import java.io.FileOutputStream;
@@ -25,7 +24,6 @@ import java.io.UnsupportedEncodingException;
 import java.lang.ref.WeakReference;
 import java.lang.reflect.Field;
 import java.lang.reflect.InvocationHandler;
-import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
 import java.lang.reflect.Proxy;
@@ -33,15 +31,18 @@ import java.net.URL;
 import java.net.URLDecoder;
 import java.nio.Buffer;
 import java.nio.ByteBuffer;
-import java.util.Arrays;
+import java.security.AccessController;
+import java.security.PrivilegedAction;
+import java.util.ArrayList;
 import java.util.Collections;
+import java.util.HashMap;
 import java.util.Iterator;
-import java.util.Map;
 import java.util.List;
+import java.util.Map;
 import java.util.WeakHashMap;
 
-import com.sun.jna.Structure.ByReference;
-
+import com.sun.jna.Callback.UncaughtExceptionHandler;
+import com.sun.jna.Structure.FFIType;
 
 /** Provides generation of invocation plumbing for a defined native
  * library interface.  Also provides various utilities for native operations.
@@ -74,6 +75,14 @@ public final class Native {
     private static Map alignments = new WeakHashMap();
     private static Map options = new WeakHashMap();
     private static Map libraries = new WeakHashMap();
+    private static final UncaughtExceptionHandler DEFAULT_HANDLER = 
+        new UncaughtExceptionHandler() {
+            public void uncaughtException(Callback c, Throwable e) {
+                System.err.println("JNA: Callback " + c + " threw the following exception:");
+                e.printStackTrace();
+            }
+        };
+    private static UncaughtExceptionHandler callbackExceptionHandler = DEFAULT_HANDLER;
     
     /** The size of a native pointer (<code>void*</code>) on the current 
      * platform, in bytes. 
@@ -83,16 +92,20 @@ public final class Native {
     public static final int LONG_SIZE;
     /** Size of a native <code>wchar_t</code> type, in bytes. */
     public static final int WCHAR_SIZE;
+    /** Size of a native <code>size_t</code> type, in bytes. */
+    public static final int SIZE_T_SIZE;
+
+    private static final int TYPE_VOIDP = 0;
+    private static final int TYPE_LONG = 1;
+    private static final int TYPE_WCHAR_T = 2;
+    private static final int TYPE_SIZE_T = 3;
+
     static {
-        try {
-            System.loadLibrary("jnidispatch");
-        }
-        catch(UnsatisfiedLinkError e) {
-            loadNativeLibrary();
-        }
-        POINTER_SIZE = pointerSize();
-        LONG_SIZE = longSize();
-        WCHAR_SIZE = wideCharSize();
+        loadNativeLibrary();
+        POINTER_SIZE = sizeof(TYPE_VOIDP);
+        LONG_SIZE = sizeof(TYPE_LONG);
+        WCHAR_SIZE = sizeof(TYPE_WCHAR_T);
+        SIZE_T_SIZE = sizeof(TYPE_SIZE_T);
         // Perform initialization of other JNA classes until *after* 
         // initializing the above final fields
         initIDs();
@@ -112,7 +125,13 @@ public final class Native {
      * system property <code>jna.protected</code> has a value of "true"
      * when the JNA library is first loaded.<p>
      * If not supported by the underlying platform, this setting will
-     * have no effect.
+     * have no effect.<p>
+     * NOTE: When protected mode is enabled, you should make use of the jsig
+     * library, if available (see <a href="http://java.sun.com/j2se/1.4.2/docs/guide/vm/signal-chaining.html">Signal Chaining</a>).
+     * In short, set the environment vaiable <code>LD_PRELOAD</code> to the
+     * path to <code>libjsig.so</code> in your JRE lib directory
+     * (usually ${java.home}/lib/${os.arch}/libjsig.so) before launching your
+     * Java application.
      */
     public static synchronized native void setProtected(boolean enable);
     
@@ -280,20 +299,17 @@ public final class Native {
      * Returns whether an instance variable was instantiated. 
      * Expects that lock on libraries is already held
      */
-    private static boolean loadInstance(Class cls) {
-        if (libraries.containsKey(cls)) {
-            return true;
-        }
-        if (cls != null) {
+    private static void loadLibraryInstance(Class cls) {
+        if (cls != null && !libraries.containsKey(cls)) {
             try {
                 Field[] fields = cls.getFields();
                 for (int i=0;i < fields.length;i++) {
                     Field field = fields[i];
                     if (field.getType() == cls 
                         && Modifier.isStatic(field.getModifiers())) {
-                        // Ensure the field gets initialized
+                        // Ensure the field gets initialized by reading it
                         libraries.put(cls, new WeakReference(field.get(null)));
-                        return true;
+                        break;
                     }
                 }
             }
@@ -302,25 +318,51 @@ public final class Native {
                                                    + cls + " (" + e + ")");
             }
         }
-        return false;
+    }
+    
+    /** Find the first instance of an interface which implements the Callback
+     * interface or an interface derived from Callback.
+     */
+    static Class findCallbackClass(Class type) {
+        if (!Callback.class.isAssignableFrom(type)) {
+            throw new IllegalArgumentException(type.getName() + " is not derived from com.sun.jna.Callback");
+        }
+        if (type.isInterface()) {
+            return type;
+        }
+        Class[] ifaces = type.getInterfaces();
+        for (int i=0;i < ifaces.length;i++) {
+            if (Callback.class.isAssignableFrom(ifaces[i])) {
+                if (ifaces[i].getMethods().length == 1)
+                    return ifaces[i];
+                break;
+            }
+        }
+        if (Callback.class.isAssignableFrom(type.getSuperclass())) {
+            return findCallbackClass(type.getSuperclass());
+        }
+        return type;
     }
     
     /** Find the library interface corresponding to the given class.  Checks
-     * all ancestor classes for a declaring class which implements 
-     * {@link Library}.
+     * all ancestor classes and interfaces for a declaring class which
+     * implements {@link Library}.
      */
-    static Class findLibraryClass(Class cls) {
+    static Class findEnclosingLibraryClass(Class cls) {
         if (cls == null) { 
             return null;
         }
         if (Library.class.isAssignableFrom(cls)) {
             return cls;
         }
-        Class fromDeclaring = findLibraryClass(cls.getDeclaringClass());
+        if (Callback.class.isAssignableFrom(cls)) {
+            cls = findCallbackClass(cls);
+        }
+        Class fromDeclaring = findEnclosingLibraryClass(cls.getDeclaringClass());
         if (fromDeclaring != null) {
             return fromDeclaring;
         }
-        return findLibraryClass(cls.getSuperclass());
+        return findEnclosingLibraryClass(cls.getSuperclass());
     }
     
 
@@ -330,13 +372,15 @@ public final class Native {
      */
     public static Map getLibraryOptions(Class type) {
         synchronized(libraries) {
-            Class interfaceClass = findLibraryClass(type);
-            if (interfaceClass == null) 
-                return null;
-            if (!loadInstance(interfaceClass)
-                || !options.containsKey(interfaceClass)) {
+            Class interfaceClass = findEnclosingLibraryClass(type);
+            if (interfaceClass != null) 
+                loadLibraryInstance(interfaceClass);
+            else
+                interfaceClass = type;
+            if (!options.containsKey(interfaceClass)) {
                 try {
                     Field field = interfaceClass.getField("OPTIONS");
+                    field.setAccessible(true);
                     options.put(interfaceClass, field.get(null));
                 }
                 catch (NoSuchFieldException e) {
@@ -355,13 +399,16 @@ public final class Native {
      */
     public static TypeMapper getTypeMapper(Class cls) {
         synchronized(libraries) {
-            Class interfaceClass = findLibraryClass(cls);
-            if (interfaceClass == null)
-                return null;
-            if (!loadInstance(interfaceClass) 
-                || !typeMappers.containsKey(interfaceClass)) {
+            Class interfaceClass = findEnclosingLibraryClass(cls);
+            if (interfaceClass != null)
+                loadLibraryInstance(interfaceClass);
+            else
+                interfaceClass = cls;
+
+            if (!typeMappers.containsKey(interfaceClass)) {
                 try {
                     Field field = interfaceClass.getField("TYPE_MAPPER");
+                    field.setAccessible(true);
                     typeMappers.put(interfaceClass, field.get(null));
                 }
                 catch (NoSuchFieldException e) {
@@ -386,17 +433,19 @@ public final class Native {
      */
     public static int getStructureAlignment(Class cls) {
         synchronized(libraries) {
-            Class interfaceClass = findLibraryClass(cls);
-            if (interfaceClass == null)
-                return Structure.ALIGN_DEFAULT;
-            if (!loadInstance(interfaceClass) 
-                || !alignments.containsKey(interfaceClass)) {
+            Class interfaceClass = findEnclosingLibraryClass(cls);
+            if (interfaceClass != null) 
+                loadLibraryInstance(interfaceClass);
+            else
+                interfaceClass = cls;
+            if (!alignments.containsKey(interfaceClass)) {
                 try {
                     Field field = interfaceClass.getField("STRUCTURE_ALIGNMENT");
+                    field.setAccessible(true);
                     alignments.put(interfaceClass, field.get(null));
                 }
                 catch(NoSuchFieldException e) {
-                    Map options = getLibraryOptions(cls);
+                    Map options = getLibraryOptions(interfaceClass);
                     if (options != null
                         && options.containsKey(Library.OPTION_STRUCTURE_ALIGNMENT)) {
                         alignments.put(interfaceClass, options.get(Library.OPTION_STRUCTURE_ALIGNMENT));
@@ -428,6 +477,24 @@ public final class Native {
         return s.getBytes();
     }
 
+    /** Obtain a NUL-terminated byte buffer equivalent to the given String. */
+    public static byte[] toByteArray(String s) {
+        byte[] bytes = getBytes(s);
+        byte[] buf = new byte[bytes.length+1];
+        System.arraycopy(bytes, 0, buf, 0, bytes.length);
+        return buf;
+    }
+
+    /** Obtain a NUL-terminated wide character buffer equivalent to the given
+        String.
+    */ 
+    public static char[] toCharArray(String s) {
+        char[] chars = s.toCharArray();
+        char[] buf = new char[chars.length+1];
+        System.arraycopy(chars, 0, buf, 0, chars.length);
+        return buf;
+    }
+
     private static String getNativeLibraryResourcePath() {
         String arch = System.getProperty("os.arch").toLowerCase();
         String osPrefix;
@@ -460,7 +527,58 @@ public final class Native {
         return "/com/sun/jna/" + osPrefix;
     }
 
+    /**
+     * Loads the JNA stub library.  It will first attempt to load this library
+     * from the directories specified in jna.boot.library.path.  If that fails,
+     * it will fallback to loading from the system library paths. Finally it will
+     * attempt to extract the stub library from from the JNA jar file, and load it.
+     * <p>
+     * The jna.boot.library.path property is mainly to support jna.jar being
+     * included in -Xbootclasspath, where java.library.path and LD_LIBRARY_PATH
+     * are ignored.  It might also be useful in other situations.
+     * </p>
+     */
     private static void loadNativeLibrary() {
+        String libName = "jnidispatch";
+        String bootPath = System.getProperty("jna.boot.library.path");
+        if (bootPath != null) {
+            String[] dirs = bootPath.split(File.pathSeparator);
+            for (int i = 0; i < dirs.length; ++i) {
+                String path = new File(new File(dirs[i]), System.mapLibraryName(libName)).getAbsolutePath();
+                try {
+                    System.load(path);
+                    return;
+                } catch (UnsatisfiedLinkError ex) {
+                }
+                if (Platform.isMac()) {
+                    String orig, ext;
+                    if (path.endsWith("dylib")) {
+                        orig = "dylib";
+                        ext = "jnilib";
+                    } else {
+                        orig = "jnilib";
+                        ext = "dylib";
+                    }
+                    try {
+                        System.load(path.substring(0, path.lastIndexOf(orig)) + ext);
+                        return;
+                    } catch (UnsatisfiedLinkError ex) {
+                    }
+                }
+            }
+        }
+        try {
+            System.loadLibrary(libName);
+        }
+        catch(UnsatisfiedLinkError e) {
+            loadNativeLibraryFromJar();
+        }
+    }
+
+    /**
+     * Extracts and loads the JNA stub library from jna.jar
+     */
+    private static void loadNativeLibraryFromJar() {
         String libname = System.mapLibraryName("jnidispatch");
         String resourceName = getNativeLibraryResourcePath() + "/" + libname;
         URL url = Native.class.getResource(resourceName);
@@ -519,14 +637,8 @@ public final class Native {
      * Initialize field and method IDs for native methods of this class. 
      * Returns the size of a native pointer.
      **/
-    private static native int pointerSize();
+    private static native int sizeof(int type);
 
-    /** Return the size of a native <code>long</code>. */
-    private static native int longSize();
-
-    /** Return the size of a native <code>wchar_t</code>. */
-    private static native int wideCharSize();
-    
     private static native String getNativeVersion();
     private static native String getAPIChecksum();
 
@@ -594,7 +706,8 @@ public final class Native {
      * <code>jna.library.path</code> so that JNA can load libraries identified
      * by the &lt;nativelib&gt; tag in the JNLP configuration file.  Returns 
      * <code>null</code> if the Web Start native library cache location can not 
-     * be determined.
+     * be determined.  Note that the path returned may be different for any
+     * given library name.
      * <p>
      * Use <code>System.getProperty("javawebstart.version")</code> to detect
      * whether your code is running under Web Start.
@@ -602,13 +715,24 @@ public final class Native {
      * Web Start class loader, which usually means it wasn't included as 
      * a <code>&lt;nativelib&gt;</code> resource in the JNLP file.
      */
-    public static String getWebStartLibraryPath(String libName) {
+    public static String getWebStartLibraryPath(final String libName) {
         if (System.getProperty("javawebstart.version") == null)
             return null;
         try {
-            ClassLoader cl = Native.class.getClassLoader();
-            Method m = ClassLoader.class.getDeclaredMethod("findLibrary", new Class[] { String.class });
-            m.setAccessible(true);
+
+            final ClassLoader cl = Native.class.getClassLoader();
+            Method m = (Method)AccessController.doPrivileged(new PrivilegedAction() {
+                public Object run() {
+                    try {
+                        Method m = ClassLoader.class.getDeclaredMethod("findLibrary", new Class[] { String.class });
+                        m.setAccessible(true);
+                        return m;
+                    }
+                    catch(Exception e) {
+                        return null;
+                    }
+                }
+            });
             String libpath = (String)m.invoke(cl, new Object[] { libName });
             if (libpath != null) {
                 return new File(libpath).getParent();
@@ -631,7 +755,7 @@ public final class Native {
      * Instead, spawn a cleanup task to remove the file *after* the VM exits.
      */
     public static class DeleteNativeLibrary extends Thread {
-        private File file;
+        private final File file;
         public DeleteNativeLibrary(File file) {
             this.file = file;
         }
@@ -742,6 +866,99 @@ public final class Native {
         }
     }
 
+    /** Set the default handler invoked when a callback throws an uncaught
+     * exception.  If the given handler is <code>null</code>, the default
+     * handler will be reinstated.
+     */
+    public static void setCallbackExceptionHandler(UncaughtExceptionHandler eh) {
+        callbackExceptionHandler = eh == null ? DEFAULT_HANDLER : eh;
+    }
+
+    /** Returns the current handler for callback uncaught exceptions. */
+    public static UncaughtExceptionHandler getCallbackExceptionHandler() {
+        return callbackExceptionHandler;
+    }
+    
+    /** When called from a class static initializer, maps all native methods
+     * found within that class to native libraries via the JNA raw calling
+     * interface.
+     * @param libName library name to which functions should be bound
+     */
+    public static void register(String libName) {
+        register(NativeLibrary.getInstance(libName));
+    }
+
+    private static Map registeredClasses = new HashMap();
+    private static Object unloader = new Object() {
+        protected void finalize() {
+            for (Iterator i=registeredClasses.entrySet().iterator();i.hasNext();) {
+                Map.Entry e = (Map.Entry)i.next();
+                unregister((Class)e.getKey(), (long[])e.getValue());
+            }
+        }
+    };
+
+    /** Unregister the native methods for the given class. */
+    private static native void unregister(Class cls, long[] handles);
+
+    /** When called from a class static initializer, maps all native methods
+     * found within that class to native libraries via the JNA raw calling
+     * interface.
+     * @param lib library to which functions should be bound
+     */
+    // TODO: Pointer, NativeLong                                            
+    // TODO: structure by value                                             
+    // TODO: byref, primitive arrays                                        
+    // TODO: stdcall name mapping                                           
+    // TODO: derive library, etc. from annotations (per-class or per-method)
+    // TODO: get function name from annotation
+    // TODO: read parameter type mapping from annotations (long/native long)
+    public static void register(NativeLibrary lib) {
+        Class[] context = new SecurityManager() {
+            public Class[] getClassContext() {
+                return super.getClassContext();
+            }
+        }.getClassContext();
+        if (context.length < 4) {
+            throw new IllegalStateException("This method must be called from the static initializer of a class");
+        }
+        Class cls = context[3];
+        Method[] methods = cls.getDeclaredMethods();
+        List mlist = new ArrayList();
+        for (int i=0;i < methods.length;i++) {
+            if ((methods[i].getModifiers() & Modifier.NATIVE) != 0) {
+                mlist.add(methods[i]);
+            }
+        }
+        long[] handles = new long[mlist.size()];
+        for (int i=0;i < handles.length;i++) {
+            Method method = (Method)mlist.get(i);
+            String sig = "(";
+            Class rtype = method.getReturnType();
+            Class[] ptypes = method.getParameterTypes();
+            for (int t=0;t < ptypes.length;t++) {
+                sig += FFIType.getSignature(ptypes[t]);
+            }
+            sig += ")" + Structure.FFIType.getSignature(rtype);
+            
+            String name = method.getName();
+            FunctionMapper mapper = (FunctionMapper)lib.getOptions().get(Library.OPTION_FUNCTION_MAPPER);
+            if (mapper != null) {
+                name = mapper.getFunctionName(lib, method);
+            }
+            Function f = lib.getFunction(name);
+            // TODO: registered methods should be disposed when
+            // the containing class or NativeLibrary is GC'd
+            handles[i] = registerMethod(cls, method.getName(),
+                                        sig, f, f.callingConvention);
+        }
+        registeredClasses.put(cls, handles);
+    }
+
+    private static native long registerMethod(Class cls,
+                                              String name, String signature,
+                                              Pointer fptr, int callingConvention);
+    
     /** Prints JNA library details to the console. */
     public static void main(String[] args) {
         final String DEFAULT_TITLE = "Java Native Access (JNA)";
