@@ -1,5 +1,6 @@
 /* Copyright (c) 2007 Wayne Meissner, All Rights Reserved
- *
+ * Copyright (c) 2007, 2008, 2009 Timothy Wall, All Rights Reserved
+ * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
@@ -44,10 +45,14 @@ public class NativeLibrary {
     private final String libraryName;
     private final String libraryPath;
     private final Map functions = new HashMap();
+    private final int callingConvention;
+    private final Map options;
 
+    private static WeakReference currentProcess;
     private static final Map libraries = new HashMap();
     private static final Map searchPaths = Collections.synchronizedMap(new HashMap());
     private static final List librarySearchPath = new LinkedList();
+    private static final List userSearchPath = new LinkedList();
 
     static {
         // Force initialization of native library
@@ -55,16 +60,22 @@ public class NativeLibrary {
             throw new Error("Native library not initialized");
     }
 
-    private NativeLibrary(String libraryName, String libraryPath, long handle) {
+    private NativeLibrary(String libraryName, String libraryPath, long handle, Map options) {
         this.libraryName = getLibraryName(libraryName);
         this.libraryPath = libraryPath;
         this.handle = handle;
+        Object option = options.get(Library.OPTION_CALLING_CONVENTION);
+        int callingConvention = option instanceof Integer
+            ? ((Integer)option).intValue() : Function.C_CONVENTION;
+        this.callingConvention = callingConvention;
+        this.options = options;
+
         // Special workaround for w32 kernel32.GetLastError
         // Short-circuit the function to use built-in GetLastError access
         if (Platform.isWindows() && "kernel32".equals(this.libraryName.toLowerCase())) {
             synchronized(functions) {
                 Function f = new Function(this, "GetLastError", Function.ALT_CONVENTION) {
-                    Object invoke(Object[] args, Class returnType) {
+                    Object invoke(Object[] args, Class returnType, boolean b) {
                         return new Integer(Native.getLastError());
                     }
                 };
@@ -73,8 +84,15 @@ public class NativeLibrary {
         }
     }
 
-    private static NativeLibrary loadLibrary(String libraryName) {
-        List searchPath = new LinkedList(librarySearchPath);
+    private static NativeLibrary loadLibrary(String libraryName, Map options) {
+        List searchPath = new LinkedList();
+
+        // Append web start path, if available.  Note that this does not
+        // attempt any library name variations
+        String webstartPath = Native.getWebStartLibraryPath(libraryName);
+        if (webstartPath != null) {
+            searchPath.add(webstartPath);
+        }
 
         //
         // Prepend any custom search paths specifically for this library
@@ -85,11 +103,28 @@ public class NativeLibrary {
                 searchPath.addAll(0, customPaths);
             }
         }
+        
+        searchPath.addAll(userSearchPath);
         String libraryPath = findLibraryPath(libraryName, searchPath);
         long handle = 0;
+        //
+        // Only search user specified paths first.  This will also fall back
+        // to dlopen/LoadLibrary() since findLibraryPath returns the mapped name
+        // if it cannot find the library.
+        //
         try {
             handle = open(libraryPath);
         }
+        catch(UnsatisfiedLinkError e) {
+            // Add the system paths back for all fallback searching
+            searchPath.addAll(librarySearchPath);
+        }
+        try {
+            if (handle == 0) {
+                libraryPath = findLibraryPath(libraryName, searchPath);
+                handle = open(libraryPath);
+            }
+        }
         catch(UnsatisfiedLinkError e) {
             if (Platform.isLinux()) {
                 //
@@ -110,12 +145,18 @@ public class NativeLibrary {
                     catch(UnsatisfiedLinkError e2) { e = e2; }
                 }
             }
+            // Try the same library with a "lib" prefix
+            else if (Platform.isWindows()) {
+                libraryPath = findLibraryPath("lib" + libraryName, searchPath);
+                try { handle = open(libraryPath); }
+                catch(UnsatisfiedLinkError e2) { e = e2; }
+            }
             if (handle == 0) {
                 throw new UnsatisfiedLinkError("Unable to load library '" + libraryName + "': "
                                                + e.getMessage());
             }
         }
-        return new NativeLibrary(libraryName, libraryPath, handle);
+        return new NativeLibrary(libraryName, libraryPath, handle, options);
     }
 
     private String getLibraryName(String libraryName) {
@@ -147,6 +188,24 @@ public class NativeLibrary {
      *      the full path to the library (e.g. "/lib/libc.so.6").
      */
     public static final NativeLibrary getInstance(String libraryName) {
+        return getInstance(libraryName, Collections.EMPTY_MAP);
+    }
+    
+    /**
+     * Returns an instance of NativeLibrary for the specified name.
+     * The library is loaded if not already loaded.  If already loaded, the
+     * existing instance is returned.<p>
+     * More than one name may map to the same NativeLibrary instance; only
+     * a single instance will be provided for any given unique file path.
+     *
+     * @param libraryName The library name to load.
+     *      This can be short form (e.g. "c"),
+     *      an explicit version (e.g. "libc.so.6"), or
+     *      the full path to the library (e.g. "/lib/libc.so.6").
+     * @param options native library options for the given library (see {@link
+     * Library}).
+     */
+    public static final NativeLibrary getInstance(String libraryName, Map options) {
         if (libraryName == null)
             throw new NullPointerException("Library name may not be null");
 
@@ -154,7 +213,7 @@ public class NativeLibrary {
             WeakReference ref = (WeakReference)libraries.get(libraryName);
             NativeLibrary library = ref != null ? (NativeLibrary)ref.get() : null;
             if (library == null) {
-                library = loadLibrary(libraryName);
+                library = loadLibrary(libraryName, options);
                 ref = new WeakReference(library);
                 libraries.put(library.getName(), ref);
                 libraries.put(library.getFile().getAbsolutePath(), ref);
@@ -164,6 +223,27 @@ public class NativeLibrary {
         }
     }
 
+    /**
+     * Returns an instance of NativeLibrary which refers to the current process.
+     * This is useful for accessing functions which were already mapped by some
+     * other mechanism, without having to reference or even know the exact
+     * name of the native library.
+     */
+    public static synchronized final NativeLibrary getProcess() {
+        NativeLibrary library = null;
+        if (currentProcess != null) {
+            library = (NativeLibrary) currentProcess.get();
+        }
+
+        if (library == null) {
+           long handle = open(null);
+           library = new NativeLibrary("<process>", null, handle, Collections.EMPTY_MAP);
+           currentProcess = new WeakReference(library);
+        }
+
+        return library;
+    }
+
     /**
      * Add a path to search for the specified library, ahead of any system paths
      *
@@ -194,7 +274,7 @@ public class NativeLibrary {
      * @throws   UnsatisfiedLinkError if the function is not found
      */
     public Function getFunction(String functionName) {
-        return getFunction(functionName, Function.C_CONVENTION);
+        return getFunction(functionName, callingConvention);
     }
 
     /**
@@ -223,6 +303,11 @@ public class NativeLibrary {
         }
     }
 
+    /** Returns this native library instance's options. */
+    public Map getOptions() {
+        return options;
+    }
+
     /** Look up the given global variable within this library.
      * @param symbolName
      * @return Pointer representing the global variable address
@@ -256,9 +341,13 @@ public class NativeLibrary {
     public String getName() {
         return libraryName;
     }
-    /** Returns the file on disk corresponding to this NativeLibrary instacne.
+    /** 
+     * Returns the file on disk corresponding to this NativeLibrary instance.
+     * If this NativeLibrary represents the current process, this function will return null. 
      */
     public File getFile() {
+        if (libraryPath == null)
+            return null;
         return new File(libraryPath);
     }
     /** Close the library when it is no longer referenced. */
@@ -269,8 +358,11 @@ public class NativeLibrary {
     public void dispose() {
         synchronized(libraries) {
             libraries.remove(getName());
-            libraries.remove(getFile().getAbsolutePath());
-            libraries.remove(getFile().getName());
+            File path = getFile();
+            if (path != null) {
+                libraries.remove(path.getAbsolutePath());
+                libraries.remove(path.getName());
+            }
         }
         synchronized(this) {
             if (handle != 0) {
@@ -343,7 +435,7 @@ public class NativeLibrary {
             return name;
         }
         else if (Platform.isLinux()) {
-            if (isVersionedName(libName)) {
+            if (isVersionedName(libName) || libName.endsWith(".so")) {
                 // A specific version was requested - use as is for search
                 return libName;
             }
@@ -404,7 +496,7 @@ public class NativeLibrary {
             String path = ((File) it.next()).getAbsolutePath();
             String ver = path.substring(path.lastIndexOf(".so.") + 4);
             double version = parseVersion(ver);
-            if (version >= bestVersion) {
+            if (version > bestVersion) {
                 bestVersion = version;
                 bestMatch = path;
             }
@@ -417,23 +509,23 @@ public class NativeLibrary {
     	double divisor = 1;
     	int dot = ver.indexOf(".");
     	while (ver != null) {
-    		String num;
-    		if (dot != -1) {
-    			num = ver.substring(0, dot);
-    			ver = ver.substring(dot + 1);
-        		dot = ver.indexOf(".");
-    		}
-    		else {
-    			num = ver;
-    			ver = null;
-    		}
-    		try {
-    			v += Integer.parseInt(num) / divisor;
-    		}
-    		catch(NumberFormatException e) {
-    			return 0;
-    		}
-    		divisor *= 100;
+            String num;
+            if (dot != -1) {
+                num = ver.substring(0, dot);
+                ver = ver.substring(dot + 1);
+                dot = ver.indexOf(".");
+            }
+            else {
+                num = ver;
+                ver = null;
+            }
+            try {
+                v += Integer.parseInt(num) / divisor;
+            }
+            catch(NumberFormatException e) {
+                return 0;
+            }
+            divisor *= 100;
     	}
 
     	return v;
@@ -443,8 +535,7 @@ public class NativeLibrary {
     private static native void close(long handle);
     private static native long findSymbol(long handle, String name);
     static {
-
-        librarySearchPath.addAll(initPaths("jna.library.path"));
+        userSearchPath.addAll(initPaths("jna.library.path"));
         String webstartPath = Native.getWebStartLibraryPath("jnidispatch");
         if (webstartPath != null) {
             librarySearchPath.add(webstartPath);
@@ -477,6 +568,13 @@ public class NativeLibrary {
                 "/usr/lib",
                 "/lib",
             };
+            // Linux 64-bit does not use /lib or /usr/lib
+            if (Platform.isLinux() && Pointer.SIZE == 8) {
+                paths = new String[] {
+                    "/usr/lib" + archPath,
+                    "/lib" + archPath,
+                };
+            }
             for (int i=0;i < paths.length;i++) {
                 File dir = new File(paths[i]);
                 if (dir.exists() && dir.isDirectory()) {
