@@ -13,6 +13,8 @@ package com.sun.jna;
 import java.io.UnsupportedEncodingException;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
+import java.util.ArrayList;
+import java.util.List;
 
 /**
  * An abstraction for a native pointer data type.  A Pointer instance 
@@ -50,7 +52,7 @@ public class Pointer {
     
     /** Pointer value of the real native pointer. Use long to be 64-bit safe. 
      */
-    long peer;
+    protected long peer;
 
     /** Derived class must assign peer pointer value. */
     Pointer() { }
@@ -66,6 +68,7 @@ public class Pointer {
     
     /** Provide a view of this pointer with a different peer base. */
     public Pointer share(long offset, long sz) {
+        if (offset == 0) return this;
         return new Pointer(peer + offset);
     }
 
@@ -603,12 +606,53 @@ v     * @param wide whether to convert from a wide or standard C string
         return buf;
     }
 
+    /** Returns an array of {@link Pointer}.  The array length is
+     * determined by a NULL-valued terminating element.
+     */
+    public Pointer[] getPointerArray(long base) {
+        List array = new ArrayList();
+        int offset = 0;
+        Pointer p = getPointer(base);
+        while (p != null) {
+            array.add(p);
+            offset += Pointer.SIZE;
+            p = getPointer(base + offset);
+        }
+        return (Pointer[])array.toArray(new Pointer[array.size()]);
+    }
+
+    /** Returns an array of {@link Pointer} of the requested size. */
     public Pointer[] getPointerArray(long offset, int arraySize) {
         Pointer[] buf = new Pointer[arraySize];
         read(offset, buf, 0, arraySize);
         return buf;
     }
 
+    /** Returns an array of <code>String</code> based on a native array
+     * of <code>char *</code>.  The array length is determined by a
+     * NULL-valued terminating element. 
+     */
+    public String[] getStringArray(long base) {
+        return getStringArray(base, false);
+    }
+
+    /** Returns an array of <code>String</code> based on a native array
+     * of <code>char*</code> or <code>wchar_t*</code> based on the
+     * <code>wide</code> parameter.  The array length is determined by a
+     * NULL-valued terminating element. 
+     */
+    public String[] getStringArray(long base, boolean wide) {
+        List strings = new ArrayList();
+        int offset = 0;
+        Pointer p = getPointer(base);
+        while (p != null) {
+            strings.add(p.getString(0, wide));
+            offset += SIZE;
+            p = getPointer(base + offset);
+        }
+        return (String[])strings.toArray(new String[strings.size()]);
+    }
+
     //////////////////////////////////////////////////////////////////////////
     // Java type write methods
     //////////////////////////////////////////////////////////////////////////
@@ -622,7 +666,7 @@ v     * @param wide whether to convert from a wide or standard C string
         _setMemory(peer + offset, length, value);
     }
     
-    private static native void _setMemory(long addr, long length, byte value);
+    static native void _setMemory(long addr, long length, byte value);
     
     /**
      * Set <code>value</code> at location being pointed to. This is equivalent
@@ -806,7 +850,7 @@ v     * @param wide whether to convert from a wide or standard C string
     /** Pointer which disallows all read/write access. */
     private static class Opaque extends Pointer {
         private Opaque(long peer) { super(peer); }
-        private String MSG = "This pointer is opaque: " + this;
+        private final String MSG = "This pointer is opaque: " + this;
         public long indexOf(long offset, byte value) {
             throw new UnsupportedOperationException(MSG);
         }
