@@ -14,10 +14,13 @@ import java.lang.reflect.Array;
 import java.lang.reflect.Field;
 import java.lang.reflect.Modifier;
 import java.nio.Buffer;
+import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.Iterator;
 import java.util.LinkedHashMap;
+import java.util.List;
 import java.util.Map;
 import java.util.Set;
 import java.util.WeakHashMap;
@@ -56,6 +59,7 @@ import java.util.WeakHashMap;
  * @author twall@users.sf.net
  */
 public abstract class Structure {
+    
     /** Tagging interface to indicate the value of an instance of the
      * <code>Structure</code> type is to be used in function invocations rather
      * than its address.  The default behavior is to treat
@@ -77,17 +81,16 @@ public abstract class Structure {
     }
 
     private static final boolean REVERSE_FIELDS;
+    static boolean REQUIRES_FIELD_ORDER;
+
     static final boolean isPPC;
     static final boolean isSPARC;
 
-
     static {
         // IBM and JRockit store fields in reverse order; check for it
         Field[] fields = MemberOrder.class.getFields();
         REVERSE_FIELDS = "last".equals(fields[0].getName());
-        if (!"middle".equals(fields[1].getName())) {
-            throw new Error("This VM does not store fields in a predictable order");
-        }
+        REQUIRES_FIELD_ORDER = !"middle".equals(fields[1].getName());
         String arch = System.getProperty("os.arch").toLowerCase();
         isPPC = "ppc".equals(arch) || "powerpc".equals(arch);
         isSPARC = "sparc".equals(arch);
@@ -102,7 +105,7 @@ public abstract class Structure {
     /** validated for w32/msvc; align on field size */
     public static final int ALIGN_MSVC = 3;
 
-    private static final int MAX_GNUC_ALIGNMENT = isSPARC ? 8 : NativeLong.SIZE;
+    private static final int MAX_GNUC_ALIGNMENT = isSPARC ? 8 : Native.LONG_SIZE;
     protected static final int CALCULATE_SIZE = -1;
 
     // This field is accessed by native code
@@ -116,19 +119,36 @@ public abstract class Structure {
     private TypeMapper typeMapper;
     // This field is accessed by native code
     private long typeInfo;
+    private List fieldOrder;
+    private boolean autoRead = true;
+    private boolean autoWrite = true;
 
     protected Structure() {
-        this(CALCULATE_SIZE);
+        this((Pointer)null);
     }
 
-    protected Structure(int size) {
-        this(size, ALIGN_DEFAULT);
+    protected Structure(TypeMapper mapper) {
+        this((Pointer)null, ALIGN_DEFAULT, mapper);
     }
 
-    protected Structure(int size, int alignment) {
+    /** Create a structure cast onto pre-allocated memory. */
+    protected Structure(Pointer p) {
+        this(p, ALIGN_DEFAULT);
+    }
+
+    protected Structure(Pointer p, int alignment) {
+        this(p, alignment, null);
+    }
+
+    protected Structure(Pointer p, int alignment, TypeMapper mapper) {
         setAlignType(alignment);
-        setTypeMapper(null);
-        allocateMemory(size);
+        setTypeMapper(mapper);
+        if (p != null) {
+            useMemory(p);
+        }
+        else {
+            allocateMemory(CALCULATE_SIZE);
+        }
     }
 
     /** Return all fields in this structure (ordered). */
@@ -150,7 +170,9 @@ public abstract class Structure {
         }
         this.typeMapper = mapper;
         this.size = CALCULATE_SIZE;
-        this.memory = null;
+        if (this.memory instanceof AutoAllocated) {
+            this.memory = null;
+        }
     }
 
     /** Change the alignment of this structure.  Re-allocates memory if
@@ -171,7 +193,9 @@ public abstract class Structure {
         }
         this.alignType = alignType;
         this.size = CALCULATE_SIZE;
-        this.memory = null;
+        if (this.memory instanceof AutoAllocated) {
+            this.memory = null;
+        }
     }
 
     /** Set the memory used by this structure.  This method is used to
@@ -189,7 +213,14 @@ public abstract class Structure {
      * memory.
      */
     protected void useMemory(Pointer m, int offset) {
-        this.memory = m.share(offset, size());
+        // Invoking size() here is important when this method is invoked
+        // from the ctor, to ensure fields are properly scanned and allocated
+        try {
+            this.memory = m.share(offset, size());
+        }
+        catch(IndexOutOfBoundsException e) {
+            throw new IllegalArgumentException("Structure exceeds provided memory bounds");
+        }
     }
 
     protected void ensureAllocated() {
@@ -207,6 +238,8 @@ public abstract class Structure {
 
     /** Provided for derived classes to indicate a different
      * size than the default.  Returns whether the operation was successful.
+     * Will leave memory untouched if it is non-null and not allocated
+     * by this class.
      */
     protected void allocateMemory(int size) {
         if (size == CALCULATE_SIZE) {
@@ -219,14 +252,13 @@ public abstract class Structure {
         // May need to defer size calculation if derived class not fully
         // initialized
         if (size != CALCULATE_SIZE) {
-            memory = new Memory(size);
-            // Always clear new structure memory
-            memory.clear(size);
-            this.size = size;
-            // Update native FFI type information, if needed
-            if (this instanceof ByValue) {
-                getTypeInfo();
+            if (this.memory == null 
+                || this.memory instanceof AutoAllocated) {
+                this.memory = new AutoAllocated(size);
+                // Always clear new structure memory
+                this.memory.clear(size);
             }
+            this.size = size;
         }
     }
 
@@ -285,11 +317,13 @@ public abstract class Structure {
         }
     }
 
-    /** Force a read of the given field from native memory.
+    /** Force a read of the given field from native memory.  The Java field
+     * will be updated from the current contents of native memory.
      * @return the new field value, after updating
      * @throws IllegalArgumentException if no field exists with the given name
      */
     public Object readField(String name) {
+        ensureAllocated();
         StructField f = (StructField)structFields.get(name);
         if (f == null)
             throw new IllegalArgumentException("No such field: " + name);
@@ -310,7 +344,7 @@ public abstract class Structure {
         }
     }
 
-    private void setField(StructField structField, Object value) {
+    void setField(StructField structField, Object value) {
         try {
             structField.field.set(this, value);
         }
@@ -321,8 +355,8 @@ public abstract class Structure {
         }
     }
 
-    /** Only keep the original structure if its address is unchanged.  Otherwise
-     * replace it with a new object.
+    /** Only keep the original structure if its native address is unchanged.
+     * Otherwise replace it with a new object.
      * @param type Structure subclass
      * @param s Original Structure object
      * @param address the native <code>struct *</code>
@@ -337,12 +371,16 @@ public abstract class Structure {
                 s = newInstance(type);
                 s.useMemory(address);
             }
-            s.read();
+            if (s.getAutoRead()) {
+                s.read();
+            }
         }
         return s;
     }
 
-    /** Read the given field and return its value. */
+    /** Read the given field and return its value.  The Java field will be
+     * updated from the contents of native memory.
+     */
     // TODO: make overridable method with calculated native type, offset, etc
     Object readField(StructField structField) {
 
@@ -355,9 +393,11 @@ public abstract class Structure {
         if (readConverter != null) {
             nativeType = readConverter.nativeType();
         }
-        // Get the value at the offset according to its type
+        // Get the current value only for types which might need to be preserved
         Object currentValue = (Structure.class.isAssignableFrom(nativeType)
                                || Callback.class.isAssignableFrom(nativeType)
+                               || Buffer.class.isAssignableFrom(nativeType)
+                               || Pointer.class.isAssignableFrom(nativeType)
                                || nativeType.isArray())
             ? getField(structField) : null;
         Object result = readValue(offset, nativeType, currentValue);
@@ -410,7 +450,15 @@ public abstract class Structure {
             result = new Double(memory.getDouble(offset));
         }
         else if (nativeType == Pointer.class) {
-            result = memory.getPointer(offset);
+            Pointer p = memory.getPointer(offset);
+            if (p != null) {
+                Pointer oldp = currentValue instanceof Pointer
+                    ? (Pointer)currentValue : null;
+                if (oldp == null || p.peer != oldp.peer)
+                    result = p;
+                else
+                    result = oldp;
+            }
         }
         else if (nativeType == String.class) {
             Pointer p = memory.getPointer(offset);
@@ -436,6 +484,19 @@ public abstract class Structure {
                 result = cb;
             }
         }
+        else if (Buffer.class.isAssignableFrom(nativeType)) {
+            Pointer bp = memory.getPointer(offset);
+            if (bp == null) {
+                result = null;
+            }
+            else {
+                Pointer oldbp = currentValue == null ? null
+                    : Native.getDirectBufferPointer((Buffer)currentValue);
+                if (oldbp == null || !oldbp.equals(bp)) {
+                    throw new IllegalStateException("Can't autogenerate a direct buffers on Structure read");
+                }
+            }
+        }
         else if (nativeType.isArray()) {
             result = currentValue;
             if (result == null) {
@@ -549,14 +610,32 @@ public abstract class Structure {
         }
     }
 
-    /** Write the given field to native memory. */
+    /** Write the given field to native memory.  The current value in the Java
+     * field will be translated into native memory.
+     * @throws IllegalArgumentException if no field exists with the given name
+     */
     public void writeField(String name) {
+        ensureAllocated();
         StructField f = (StructField)structFields.get(name);
         if (f == null)
             throw new IllegalArgumentException("No such field: " + name);
         writeField(f);
     }
 
+    /** Write the given field value to the field and native memory.   The
+     * given value will be written both to the Java field and the
+     * corresponding native memory.
+     * @throws IllegalArgumentException if no field exists with the given name
+     */
+    public void writeField(String name, Object value) {
+        ensureAllocated();
+        StructField f = (StructField)structFields.get(name);
+        if (f == null)
+            throw new IllegalArgumentException("No such field: " + name);
+        setField(f, value);
+        writeField(f);
+    }
+
     void writeField(StructField structField) {
         // Get the offset of the field
         int offset = structField.offset;
@@ -587,6 +666,7 @@ public abstract class Structure {
             }
             else {
                 value = null;
+                nativeStrings.remove(structField.name);
             }
         }
 
@@ -606,25 +686,25 @@ public abstract class Structure {
             memory.setInt(offset, Boolean.TRUE.equals(value) ? -1 : 0);
         }
         else if (nativeType == byte.class || nativeType == Byte.class) {
-            memory.setByte(offset, ((Byte)value).byteValue());
+            memory.setByte(offset, value == null ? 0 : ((Byte)value).byteValue());
         }
         else if (nativeType == short.class || nativeType == Short.class) {
-            memory.setShort(offset, ((Short)value).shortValue());
+            memory.setShort(offset, value == null ? 0 : ((Short)value).shortValue());
         }
         else if (nativeType == char.class || nativeType == Character.class) {
-            memory.setChar(offset, ((Character)value).charValue());
+            memory.setChar(offset, value == null ? 0 : ((Character)value).charValue());
         }
         else if (nativeType == int.class || nativeType == Integer.class) {
-            memory.setInt(offset, ((Integer)value).intValue());
+            memory.setInt(offset, value == null ? 0 : ((Integer)value).intValue());
         }
         else if (nativeType == long.class || nativeType == Long.class) {
-            memory.setLong(offset, ((Long)value).longValue());
+            memory.setLong(offset, value == null ? 0 : ((Long)value).longValue());
         }
         else if (nativeType == float.class || nativeType == Float.class) {
-            memory.setFloat(offset, ((Float)value).floatValue());
+            memory.setFloat(offset, value == null ? 0f : ((Float)value).floatValue());
         }
         else if (nativeType == double.class || nativeType == Double.class) {
-            memory.setDouble(offset, ((Double)value).doubleValue());
+            memory.setDouble(offset, value == null ? 0.0 : ((Double)value).doubleValue());
         }
         else if (nativeType == Pointer.class) {
             memory.setPointer(offset, (Pointer)value);
@@ -651,6 +731,11 @@ public abstract class Structure {
         else if (Callback.class.isAssignableFrom(nativeType)) {
             memory.setPointer(offset, CallbackReference.getFunctionPointer((Callback)value));
         }
+        else if (nativeType == Buffer.class) {
+            Pointer p = value == null ? null
+                : Native.getDirectBufferPointer((Buffer)value);
+            memory.setPointer(offset, p);
+        }
         else if (nativeType.isArray()) {
             return writeArrayValue(offset, value, nativeType.getComponentType());
         }
@@ -732,6 +817,35 @@ public abstract class Structure {
         return true;
     }
 
+    protected List getFieldOrder() {
+        synchronized(this) {
+            if (fieldOrder == null) {
+                fieldOrder = new ArrayList();
+            }
+            return fieldOrder;
+        }
+    }
+
+    /** Provided for VMs where the field order as returned by {@link
+     * Class#getFields()} is not predictable.
+     */
+    protected void setFieldOrder(String[] fields) {
+        getFieldOrder().addAll(Arrays.asList(fields));
+    }
+
+    /** Sort the structure fields according to the given array of names. */
+    protected void sortFields(Field[] fields, String[] names) {
+        for (int i=0;i < names.length;i++) {
+            for (int f=i;f < fields.length;f++) {
+                if (names[i].equals(fields[f].getName())) {
+                    Field tmp = fields[f];
+                    fields[f] = fields[i];
+                    fields[i] = tmp;
+                    break;
+                }
+            }
+        }
+    }
 
     /** Calculate the amount of native memory required for this structure.
      * May return {@link #CALCULATE_SIZE} if the size can not yet be
@@ -752,6 +866,16 @@ public abstract class Structure {
         structAlignment = 1;
         int calculatedSize = 0;
         Field[] fields = getClass().getFields();
+        // Restrict to valid fields
+        List flist = new ArrayList();
+        for (int i=0;i < fields.length;i++) {
+            int modifiers = fields[i].getModifiers();
+            if (Modifier.isStatic(modifiers) || !Modifier.isPublic(modifiers))
+                continue;
+            flist.add(fields[i]);
+        }
+        fields = (Field[])flist.toArray(new Field[flist.size()]);
+
         if (REVERSE_FIELDS) {
             for (int i=0;i < fields.length/2;i++) {
                 int idx = fields.length-1-i;
@@ -760,11 +884,20 @@ public abstract class Structure {
                 fields[idx] = tmp;
             }
         }
+        else if (REQUIRES_FIELD_ORDER) {
+            List fieldOrder = getFieldOrder();
+            if (fieldOrder.size() < fields.length) {
+                if (force) {
+                    throw new Error("This VM does not store fields in a predictable order; you must use setFieldOrder: " + System.getProperty("java.vendor") + ", " + System.getProperty("java.version"));
+                }
+                return CALCULATE_SIZE;
+            }
+            sortFields(fields, (String[])fieldOrder.toArray(new String[fieldOrder.size()]));
+        }
+
         for (int i=0; i<fields.length; i++) {
             Field field = fields[i];
             int modifiers = field.getModifiers();
-            if (Modifier.isStatic(modifiers) || !Modifier.isPublic(modifiers))
-                continue;
 
             Class type = field.getType();
             StructField structField = new StructField();
@@ -819,12 +952,14 @@ public abstract class Structure {
             Class nativeType = type;
             if (NativeMapped.class.isAssignableFrom(type)) {
                 NativeMappedConverter tc = NativeMappedConverter.getInstance(type);
-                value = tc.defaultValue();
+                if (value == null) {
+                    value = tc.defaultValue();
+                    setField(structField, value);
+                }
                 nativeType = tc.nativeType();
                 structField.writeConverter = tc;
                 structField.readConverter = tc;
                 structField.context = new StructureReadContext(this, field);
-                setField(structField, value);
             }
             else if (typeMapper != null) {
                 ToNativeConverter writeConverter = typeMapper.getToNativeConverter(type);
@@ -842,8 +977,17 @@ public abstract class Structure {
                     throw new IllegalArgumentException(msg);
                 }
             }
-            structField.size = getNativeSize(nativeType, value);
-            fieldAlignment = getNativeAlignment(nativeType, value, i==0);
+            try {
+                structField.size = getNativeSize(nativeType, value);
+                fieldAlignment = getNativeAlignment(nativeType, value, i==0);
+            }
+            catch(IllegalArgumentException e) {
+                // Might simply not yet have a type mapper set
+                if (!force && typeMapper == null) {
+                    return CALCULATE_SIZE;
+                }
+                throw e;
+            }
 
             // Align fields as appropriate
             structAlignment = Math.max(structAlignment, fieldAlignment);
@@ -858,7 +1002,12 @@ public abstract class Structure {
         }
 
         if (calculatedSize > 0) {
-            return calculateAlignedSize(calculatedSize);
+            int size = calculateAlignedSize(calculatedSize);
+            // Update native FFI type information, if needed
+            if (this instanceof ByValue) {
+                getTypeInfo();
+            }
+            return size;
         }
 
         throw new IllegalArgumentException("Structure " + getClass()
@@ -899,7 +1048,7 @@ public abstract class Structure {
         int size = getNativeSize(type, value);
         if (type.isPrimitive() || Long.class == type || Integer.class == type
             || Short.class == type || Character.class == type
-            || Byte.class == type
+            || Byte.class == type || Boolean.class == type
             || Float.class == type || Double.class == type) {
             alignment = size;
         }
@@ -927,15 +1076,18 @@ public abstract class Structure {
             throw new IllegalArgumentException("Type " + type + " has unknown "
                                                + "native alignment");
         }
-        if (alignType == ALIGN_NONE)
-            return 1;
-        if (alignType == ALIGN_MSVC)
-            return Math.min(8, alignment);
-        if (alignType == ALIGN_GNUC) {
+        if (alignType == ALIGN_NONE) {
+            alignment = 1;
+        }
+        else if (alignType == ALIGN_MSVC) {
+            alignment = Math.min(8, alignment);
+        }
+        else if (alignType == ALIGN_GNUC) {
             // NOTE this is published ABI for 32-bit gcc/linux/x86, osx/x86,
             // and osx/ppc.  osx/ppc special-cases the first element
-            if (!isFirstElement || !(Platform.isMac() && isPPC))
-                return Math.min(MAX_GNUC_ALIGNMENT, alignment);
+            if (!isFirstElement || !(Platform.isMac() && isPPC)) {
+                alignment = Math.min(MAX_GNUC_ALIGNMENT, alignment);
+            }
         }
         return alignment;
     }
@@ -951,12 +1103,6 @@ public abstract class Structure {
             // Don't process zero-length arrays
             throw new IllegalArgumentException("Arrays of length zero not allowed in structure: " + type);
         }
-        // May provide this in future; problematic on read, since we can't
-        // auto-create a java.nio.Buffer w/o knowing its size
-        if (Buffer.class.isAssignableFrom(type)) {
-            throw new IllegalArgumentException("the type \"" + type.getName()
-                                               + "\" is not supported as a structure field");
-        }
         if (Structure.class.isAssignableFrom(type)
             && !Structure.ByReference.class.isAssignableFrom(type)) {
             if (value == null)
@@ -968,7 +1114,8 @@ public abstract class Structure {
         }
         catch(IllegalArgumentException e) {
             throw new IllegalArgumentException("The type \"" + type.getName()
-                                               + "\" is not supported as a structure field");
+                                               + "\" is not supported as a structure field: "
+                                               + e.getMessage());
         }
     }
 
@@ -1046,15 +1193,13 @@ public abstract class Structure {
      * backing, the memory will be resized to fit the entire array.
      */
     public Structure[] toArray(Structure[] array) {
-        if (size == CALCULATE_SIZE) {
-            allocateMemory();
-        }
-        if (Memory.class.equals(memory.getClass())) {
+        ensureAllocated();
+        if (memory instanceof AutoAllocated) {
             // reallocate if necessary
             Memory m = (Memory)memory;
             int requiredSize = array.length * size();
             if (m.getSize() < requiredSize) {
-                m = new Memory(requiredSize);
+                m = new AutoAllocated(requiredSize);
                 m.clear();
                 useMemory(m);
             }
@@ -1109,13 +1254,58 @@ public abstract class Structure {
         return p != null ? p.hashCode() : 0;
     }
 
+    protected void cacheTypeInfo(Pointer p) {
+        typeInfo = p.peer;
+    }
+
     /** Obtain native type information for this structure. */
     Pointer getTypeInfo() {
         Pointer p = getTypeInfo(this);
-        typeInfo = p.peer;
+        cacheTypeInfo(p);
         return p;
     }
 
+    /** Set whether the structure is automatically synched to native memory
+        before and after a native function call.  Convenience method for
+        <pre><code>
+        boolean auto = ...;
+        setAutoRead(auto);
+        setAutoWrite(auto);
+        </code></pre>
+    */
+    public void setAutoSynch(boolean auto) {
+        setAutoRead(auto);
+        setAutoWrite(auto);
+    }
+
+    /** Set whether the struture is written to native memory prior to
+        a native function call.
+    */
+    public void setAutoRead(boolean auto) {
+        this.autoRead = auto;
+    }
+
+    /** Returns whether the struture is written to native memory prior to
+        a native function call.
+    */
+    public boolean getAutoRead() {
+        return this.autoRead;
+    }
+
+    /** Set whether the structure is read from native memory after a native
+        function call. 
+    */
+    public void setAutoWrite(boolean auto) {
+        this.autoWrite = auto;
+    }
+
+    /** Returns whether the structure is read from native memory after a native
+        function call. 
+    */
+    public boolean getAutoWrite() {
+        return this.autoWrite;
+    }
+
     /** Exposed for testing purposes only. */
     static Pointer getTypeInfo(Object obj) {
         return FFIType.get(obj);
@@ -1160,7 +1350,7 @@ public abstract class Structure {
      * structure for use by libffi.  The lifecycle of this structure is easier
      * to manage on the Java side than in native code.
      */
-    private static class FFIType extends Structure {
+    static class FFIType extends Structure {
         public static class size_t extends IntegerType {
             public size_t() { this(0); }
             public size_t(long value) { super(Native.POINTER_SIZE, value); }
@@ -1220,11 +1410,20 @@ public abstract class Structure {
         public Pointer elements;
 
         private FFIType(Structure ref) {
-            Pointer[] els = new Pointer[ref.fields().size() + 1];
-            int idx = 0;
-            for (Iterator i=ref.fields().values().iterator();i.hasNext();) {
-                StructField sf = (StructField)i.next();
-                els[idx++] = get(ref.getField(sf), sf.type);
+            Pointer[] els;
+            if (ref instanceof Union) {
+                StructField sf = ((Union)ref).biggestField;
+                els = new Pointer[] {
+                    get(ref.getField(sf), sf.type), null,
+                };
+            }
+            else {
+                els = new Pointer[ref.fields().size() + 1];
+                int idx = 0;
+                for (Iterator i=ref.fields().values().iterator();i.hasNext();) {
+                    StructField sf = (StructField)i.next();
+                    els[idx++] = get(ref.getField(sf), sf.type);
+                }
             }
             init(els);
         }
@@ -1247,6 +1446,8 @@ public abstract class Structure {
         static Pointer get(Object obj) {
             if (obj == null)
                 return FFITypes.ffi_type_pointer;
+            if (obj instanceof Class)
+                return get(null, (Class)obj);
             return get(obj, obj.getClass());
         }
 
@@ -1270,9 +1471,6 @@ public abstract class Structure {
                         typeInfoMap.put(cls, FFITypes.ffi_type_pointer);
                         return FFITypes.ffi_type_pointer;
                     }
-                    if (Union.class.isAssignableFrom(cls)) {
-                        return ((Union)obj).getTypeInfo();
-                    }
                     FFIType type = new FFIType((Structure)obj);
                     typeInfoMap.put(cls, type);
                     return type.getPointer();
@@ -1290,5 +1488,24 @@ public abstract class Structure {
                 throw new IllegalArgumentException("Unsupported structure field type " + cls);
             }
         }
+        static String getSignature(Object o) {
+            Pointer p = get(o);
+            if (p == FFITypes.ffi_type_void) return "V";
+            if (p == FFITypes.ffi_type_float) return "F";
+            if (p == FFITypes.ffi_type_double) return "D";
+            if (p == FFITypes.ffi_type_sint8) return "B";
+            if (p == FFITypes.ffi_type_sint16) return "S";
+            if (p == FFITypes.ffi_type_uint16) return "C";
+            if (p == FFITypes.ffi_type_sint32) return "I";
+            if (p == FFITypes.ffi_type_sint64) return "J";
+            if (p == FFITypes.ffi_type_pointer) return "Lcom/sun/jna/Pointer;";
+            throw new IllegalArgumentException("Unsupported type " + o);
+        }
+    }
+    
+    private class AutoAllocated extends Memory {
+        public AutoAllocated(int size) {
+            super(size);
+        }
     }
 }
