@@ -1,4 +1,4 @@
-/* Copyright (c) 2007 Timothy Wall, All Rights Reserved
+/* Copyright (c) 2007-2008 Timothy Wall, All Rights Reserved
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -12,10 +12,13 @@
  */
 package com.sun.jna;
 
+import java.util.HashMap;
 import java.util.Map;
+import java.lang.ref.WeakReference;
 
 import com.sun.jna.CallbacksTest.TestLibrary.CbCallback;
 import com.sun.jna.ptr.IntByReference;
+import com.sun.jna.Callback.UncaughtExceptionHandler;
 
 import junit.framework.TestCase;
 
@@ -47,6 +50,14 @@ public class CallbacksTest extends TestCase {
             void callback();
         }
         void callVoidCallback(VoidCallback c);
+        interface VoidCallbackCustom extends Callback {
+            void customMethodName();
+        }
+        abstract class VoidCallbackCustomAbstract implements VoidCallbackCustom {
+            public void customMethodName() { }
+        }
+        class VoidCallbackCustomDerived extends VoidCallbackCustomAbstract { }
+        void callVoidCallback(VoidCallbackCustom c);
         interface BooleanCallback extends Callback {
             boolean callback(boolean arg, boolean arg2);
         }
@@ -94,12 +105,22 @@ public class CallbacksTest extends TestCase {
         interface CopyArgToByReference extends Callback {
         	int callback(int arg, IntByReference result);
         }
+        interface StringArrayCallback extends Callback {
+            String[] callback(String[] arg);
+        }
+        Pointer callStringArrayCallback(StringArrayCallback c, String[] arg);
         int callCallbackWithByReferenceArgument(CopyArgToByReference cb, int arg, IntByReference result);
         TestStructure.ByValue callCallbackWithStructByValue(TestStructure.TestCallback callback, TestStructure.ByValue cbstruct);
         interface CbCallback extends Callback {
             CbCallback callback(CbCallback arg);
         }
         CbCallback callCallbackWithCallback(CbCallback cb);
+
+        public interface Int32CallbackX extends Callback {
+            public int callback(int arg);
+        }
+        Int32CallbackX returnCallback();
+        Int32CallbackX returnCallbackArgument(Int32CallbackX cb);
     }
 
     TestLibrary lib;
@@ -146,6 +167,17 @@ public class CallbacksTest extends TestCase {
         assertEquals("Callback trampoline not freed", 0, cbstruct.peer);
     }
     
+    public void testFindCallbackInterface() {
+        TestLibrary.Int32Callback cb = new TestLibrary.Int32Callback() {
+            public int callback(int arg, int arg2) {
+                return arg + arg2;
+            }
+        };
+        assertEquals("Wrong callback interface",
+                     TestLibrary.Int32Callback.class,
+                     Native.findCallbackClass(cb.getClass()));
+    }
+
     public void testCallInt32Callback() {
         final int MAGIC = 0x11111111;
         final boolean[] called = { false };
@@ -389,6 +421,32 @@ public class CallbacksTest extends TestCase {
         assertEquals("Wrong String return", VALUE, value);
     }
     
+    public void testStringCallbackMemoryReclamation() throws InterruptedException {
+        TestLibrary.StringCallback cb = new TestLibrary.StringCallback() {
+            public String callback(String arg) {
+                return arg;
+            }
+        };
+
+        // A little internal groping
+        Map m = CallbackReference.allocations;
+        String arg = getName() + "1";
+        String value = lib.callStringCallback(cb, arg);
+        WeakReference ref = new WeakReference(value);
+        
+        arg = null;
+        value = null;
+        System.gc();
+        for (int i = 0; i < 100 && (ref.get() != null || m.values().size() > 0); ++i) {
+            try {
+                Thread.sleep(10); // Give the GC a chance to run
+                System.gc();
+            } finally {}
+        }
+        assertNull("String reference not GC'd", ref.get());
+        assertEquals("NativeString reference still held", 0, m.values().size());
+    }
+
     public void testCallWideStringCallback() {
         final boolean[] called = {false};
         final WString[] cbargs = { null };
@@ -406,6 +464,24 @@ public class CallbacksTest extends TestCase {
         assertEquals("Wrong wide string return", VALUE, value);
     }
     
+    public void testCallStringArrayCallback() {
+        final boolean[] called = {false};
+        final String[][] cbargs = { null };
+        TestLibrary.StringArrayCallback cb = new TestLibrary.StringArrayCallback() {
+            public String[] callback(String[] arg) {
+                called[0] = true;
+                cbargs[0] = arg;
+                return arg;
+            }
+        };
+        final String[] VALUE = { "value", null };
+        Pointer value = lib.callStringArrayCallback(cb, VALUE);
+        assertTrue("Callback not called", called[0]);
+        assertEquals("Wrong callback argument 1", VALUE[0], cbargs[0][0]);
+        String[] result = value.getStringArray(0);
+        assertEquals("Wrong String return", VALUE[0], result[0]);
+    }
+    
     public void testCallCallbackWithByReferenceArgument() {
     	final boolean[] called = {false};
         TestLibrary.CopyArgToByReference cb = new TestLibrary.CopyArgToByReference() {
@@ -427,6 +503,7 @@ public class CallbacksTest extends TestCase {
         final TestStructure innerResult = new TestStructure();
         TestStructure.TestCallback cb = new TestStructure.TestCallback() {
             public TestStructure.ByValue callback(TestStructure.ByValue s) {
+                // Copy the argument value for later comparison
                 Pointer old = innerResult.getPointer();
                 innerResult.useMemory(s.getPointer());
                 innerResult.read();
@@ -456,6 +533,194 @@ public class CallbacksTest extends TestCase {
         assertEquals("Callback reference should be reused", cb, cb2);
     }
     
+    /* Most Callbacks are wrapped in DefaultCallbackProxy, which catches their
+     * exceptions.
+     */
+    public void testCallbackExceptionHandler() {
+        final RuntimeException ERROR = new RuntimeException(getName());
+        final Throwable CAUGHT[] = { null };
+        final Callback CALLBACK[] = { null };
+        UncaughtExceptionHandler old = Native.getCallbackExceptionHandler();
+        UncaughtExceptionHandler handler = new UncaughtExceptionHandler() {
+            public void uncaughtException(Callback cb, Throwable e) {
+                CALLBACK[0] = cb;
+                CAUGHT[0] = e;
+            }
+        };
+        Native.setCallbackExceptionHandler(handler);
+        try {
+            TestLibrary.CbCallback cb = new TestLibrary.CbCallback() {
+                public CbCallback callback(CbCallback arg) {
+                    throw ERROR;
+                }
+            };
+            TestLibrary.CbCallback cb2 = lib.callCallbackWithCallback(cb);
+            assertNotNull("Exception handler not called", CALLBACK[0]);
+            assertEquals("Wrong callback argument to handler", cb, CALLBACK[0]);
+            assertEquals("Wrong exception passed to handler",
+                         ERROR, CAUGHT[0]);
+        }
+        finally {
+            Native.setCallbackExceptionHandler(old);
+        }
+    }
+
+    /* CallbackProxy is called directly from native. */
+    public void testCallbackExceptionHandlerWithCallbackProxy() throws Throwable {
+        final RuntimeException ERROR = new RuntimeException(getName());
+        final Throwable CAUGHT[] = { null };
+        final Callback CALLBACK[] = { null };
+        UncaughtExceptionHandler old = Native.getCallbackExceptionHandler();
+        UncaughtExceptionHandler handler = new UncaughtExceptionHandler() {
+            public void uncaughtException(Callback cb, Throwable e) {
+                CALLBACK[0] = cb;
+                CAUGHT[0] = e;
+            }
+        };
+        Native.setCallbackExceptionHandler(handler);
+        try {
+            class TestProxy implements CallbackProxy, TestLibrary.CbCallback {
+                public CbCallback callback(CbCallback arg) {
+                    throw new Error("Should never be called");
+                }
+                public Object callback(Object[] args) {
+                    throw ERROR;
+                }
+                public Class[] getParameterTypes() {
+                    return new Class[] { CbCallback.class };
+                }
+                public Class getReturnType() {
+                    return CbCallback.class;
+                }
+            };
+            TestLibrary.CbCallback cb = new TestProxy();
+            TestLibrary.CbCallback cb2 = lib.callCallbackWithCallback(cb);
+            assertNotNull("Exception handler not called", CALLBACK[0]);
+            assertEquals("Wrong callback argument to handler", cb, CALLBACK[0]);
+            assertEquals("Wrong exception passed to handler",
+                         ERROR, CAUGHT[0]);
+        }
+        finally {
+            Native.setCallbackExceptionHandler(old);
+        }
+    }
+
+    public void testResetCallbackExceptionHandler() {
+        Native.setCallbackExceptionHandler(null);
+        assertNotNull("Should not be able to set callback EH null",
+                      Native.getCallbackExceptionHandler());
+    }
+
+    public static interface CallbackTestLibrary extends Library {
+        interface Int32Callback extends Callback {
+            float callback(float arg, float arg2);
+        }
+        float callInt32Callback(Int32Callback c, float arg, float arg2);
+    }
+
+    public void testCallbackTypeMappingFromLibrary() throws Exception {
+        final DefaultTypeMapper mapper = new DefaultTypeMapper();
+        Map options = new HashMap() {
+            { put(Library.OPTION_TYPE_MAPPER, mapper); }
+        };
+        CallbackTestLibrary lib = (CallbackTestLibrary)
+            Native.loadLibrary("testlib", CallbackTestLibrary.class, options);
+
+        // Convert java floats into native integers and back
+        TypeConverter converter = new TypeConverter() {
+            public Object fromNative(Object value, FromNativeContext context) {
+                return new Float(((Integer)value).intValue());
+            }
+            public Class nativeType() {
+                return Integer.class;
+            }
+            public Object toNative(Object value, ToNativeContext ctx) {
+                return new Integer(Math.round(((Float)value).floatValue()));
+            }
+        };
+        mapper.addTypeConverter(float.class, converter);
+        CallbackTestLibrary.Int32Callback cb = new CallbackTestLibrary.Int32Callback() {
+            public float callback(float arg, float arg2) {
+                return arg + arg2;
+            }
+        };
+        assertEquals("Wrong type mapper for callback class", mapper,
+                     Native.getTypeMapper(CallbackTestLibrary.Int32Callback.class));
+        assertEquals("Wrong type mapper for callback object", mapper,
+                     Native.getTypeMapper(cb.getClass()));
+
+        assertEquals("Wrong type mapper used in callback invocation",
+                     -2, lib.callInt32Callback(cb, -1, -1), 0);
+    }
+
+    private static class TestCallback implements Callback {
+        public static final TypeMapper TYPE_MAPPER = new DefaultTypeMapper();
+        public void callback() { }
+    }
+    public void testCallbackTypeMappingFromCallback() throws Exception {
+        assertEquals("Wrong type mapper for callback class",
+                     TestCallback.TYPE_MAPPER,
+                     Native.getTypeMapper(TestCallback.class));
+    }
+
+    public void testInvokeCallback() {
+        TestLibrary lib = (TestLibrary)
+            Native.loadLibrary("testlib", TestLibrary.class);
+        TestLibrary.Int32CallbackX cb = lib.returnCallback();
+        assertNotNull("Callback should not be null", cb);
+        assertEquals("Callback should be callable", 1, cb.callback(1));
+        
+        TestLibrary.Int32CallbackX cb2 = new TestLibrary.Int32CallbackX() {
+            public int callback(int arg) {
+                return 0;
+            }
+        };
+        assertSame("Java callback should be looked up",
+                   cb2, lib.returnCallbackArgument(cb2));
+        assertSame("Existing native function wrapper should be reused",
+                   cb, lib.returnCallbackArgument(cb));
+    }
+    
+    static class CbStruct extends Structure {
+        public Callback cb;
+    }
+    interface CbTest extends Library {
+        public void callCallbackInStruct(CbStruct cbstruct);
+    }
+    public void testCallCallbackInStructure() {
+        final boolean[] flag = {false};
+        final CbStruct s = new CbStruct();
+        s.cb = new Callback() {
+            public void callback() {
+                flag[0] = true;
+            }
+        };
+        CbTest lib = (CbTest)Native.loadLibrary("testlib", CbTest.class);
+        lib.callCallbackInStruct(s);
+        assertTrue("Callback not invoked", flag[0]);
+    }
+
+    public void testCustomCallbackMethodName() {
+    	final boolean[] called = {false};
+        TestLibrary.VoidCallbackCustom cb = new TestLibrary.VoidCallbackCustom() {
+            public void customMethodName() {
+                called[0] = true;
+            }
+            public String toString() {
+                return "Some debug output";
+            }
+        };
+        lib.callVoidCallback(cb);
+        assertTrue("Callback with custom method name not called", called[0]);
+    }
+
+    public void testCustomCallbackVariedInheritance() {
+    	final boolean[] called = {false};
+        TestLibrary.VoidCallbackCustom cb =
+            new TestLibrary.VoidCallbackCustomDerived();
+        lib.callVoidCallback(cb);
+    }
+
     public static void main(java.lang.String[] argList) {
         junit.textui.TestRunner.run(CallbacksTest.class);
     }
