@@ -47,6 +47,37 @@ public class StructureTest extends TestCase {
         public int f2;
         public int f3;
     }
+
+
+    public void testFieldsAllocated() {
+        class TestStructure extends Structure {
+            public TestStructure() { }
+            public TestStructure(Pointer p) { super(p); }
+            public int field;
+            public int fieldCount() { return fields().size(); }
+        }
+        TestStructure s = new TestStructure();
+        assertEquals("Wrong number of fields (default)", 1, s.fieldCount());
+
+        s = new TestStructure(new Memory(4));
+        assertEquals("Wrong number of fields (preallocated)", 1, s.fieldCount());
+    }
+
+    public void testProvidedMemoryTooSmall() {
+        class TestStructure extends Structure {
+            public TestStructure() { }
+            public TestStructure(Pointer p) { super(p); }
+            public int field;
+            public int fieldCount() { return fields().size(); }
+        }
+        try {
+            TestStructure s = new TestStructure(new Memory(2));
+            fail("Expect exception if provided memory is insufficient");
+        }
+        catch(IllegalArgumentException e) {
+        }
+    }
+
     public void testClearOnAllocate() {
         TestAllocStructure s = new TestAllocStructure();
         s.read();
@@ -218,7 +249,18 @@ public class StructureTest extends TestCase {
         assertEquals("Wrong aggregate size",
                      s.s1.size() + s.s2.size() + 4, s.size());
         s.write();
+        assertEquals("Wrong memory for structure field 1 after write",
+                     s.getPointer(), s.s1.getPointer());
+        assertEquals("Wrong memory for structure field 2 after write",
+                     s.getPointer().share(s.s1.size()),
+                     s.s2.getPointer());
+
         s.read();
+        assertEquals("Wrong memory for structure field 1 after read",
+                     s.getPointer(), s.s1.getPointer());
+        assertEquals("Wrong memory for structure field 2 after read",
+                     s.getPointer().share(s.s1.size()),
+                     s.s2.getPointer());
     }
 
     public void testPrimitiveArrayField() {
@@ -419,7 +461,7 @@ public class StructureTest extends TestCase {
             public Function cb;
         }
         try {
-            BadFieldStructure s = new BadFieldStructure();
+            new BadFieldStructure().size();
             fail("Function fields should not be allowed");
         }
         catch(IllegalArgumentException e) {
@@ -445,7 +487,6 @@ public class StructureTest extends TestCase {
         public TestCallback cb;
     }
     static interface CbTest extends Library {
-        public void callCallbackInStruct(CbStruct cbstruct);
         public void setCallbackInStruct(CbStruct2 cbstruct);
     }
     public void testCallbackWrite() {
@@ -463,19 +504,6 @@ public class StructureTest extends TestCase {
         assertEquals("Wrong trampoline", ref.getTrampoline(), func);
     }
 
-    public void testCallCallbackInStructure() {
-        final boolean[] flag = {false};
-        final CbStruct s = new CbStruct();
-        s.cb = new Callback() {
-            public void callback() {
-                flag[0] = true;
-            }
-        };
-        CbTest lib = (CbTest)Native.loadLibrary("testlib", CbTest.class);
-        lib.callCallbackInStruct(s);
-        assertTrue("Callback not invoked", flag[0]);
-    }
-
     public void testReadFunctionPointerAsCallback() {
         CbStruct2 s = new CbStruct2();
         CbTest lib = (CbTest)Native.loadLibrary("testlib", CbTest.class);
@@ -534,20 +562,51 @@ public class StructureTest extends TestCase {
         assertEquals("Wrong first element", s.getPointer(), s.array[0]);
     }
 
-    public void testBufferField() {
-        // NOTE: may support write-only Buffer fields in the future
-        class BufferStructure extends Structure {
-            public Buffer buffer;
-            public BufferStructure(byte[] buf) {
-                buffer = ByteBuffer.wrap(buf);
-            }
+    class BufferStructure extends Structure {
+        public Buffer buffer;
+    }
+    public void testBufferFieldWriteNULL() {
+        BufferStructure bs = new BufferStructure();
+        bs.write();
+    }
+    public void testBufferFieldWriteNonNULL() {
+        BufferStructure bs = new BufferStructure();
+        bs.buffer = ByteBuffer.allocateDirect(16);
+        bs.write();
+    }
+    public void testBufferFieldReadUnchanged() {
+        BufferStructure bs = new BufferStructure();
+        bs.buffer = ByteBuffer.allocateDirect(16);
+        bs.write();
+        bs.read();
+    }
+    public void testBufferFieldReadChanged() {
+        BufferStructure bs = new BufferStructure();
+        if (Pointer.SIZE == 4) {
+            bs.getPointer().setInt(0, 0x1);
         }
-    	try {
-    		new BufferStructure(new byte[1024]);
-    		fail("Buffer fields should fail immediately");
-    	}
-    	catch(IllegalArgumentException e) {
-    	}
+        else {
+            bs.getPointer().setLong(0, 0x1);
+        }
+        try {
+            bs.read();
+            fail("Structure read should fail if Buffer pointer was set");
+        }
+        catch(IllegalStateException e) {
+        }
+        bs.buffer = ByteBuffer.allocateDirect(16);
+        try {
+            bs.read();
+            fail("Structure read should fail if Buffer pointer has changed");
+        }
+        catch(IllegalStateException e) {
+        }
+    }
+    public void testBufferFieldReadChangedToNULL() {
+        BufferStructure bs = new BufferStructure();
+        bs.buffer = ByteBuffer.allocateDirect(16);
+        bs.read();
+        assertNull("Structure Buffer field should be set null", bs.buffer);
     }
 
     public void testVolatileStructureField() {
@@ -600,6 +659,21 @@ public class StructureTest extends TestCase {
         assertTrue("Read should preserve structure memory",
                    inner.getPointer() instanceof Memory);
     }
+    
+    public static class TestPointer extends PointerType { }
+    public void testPreservePointerFields() {
+        class TestStructure extends Structure {
+            public Pointer p = new Memory(256);
+            public TestPointer p2 = new TestPointer() {
+                    { setPointer(new Memory(256)); }
+                };
+        }
+        TestStructure s = new TestStructure();
+        final Pointer p = s.p;
+        s.write();
+        s.read();
+        assertSame("Should preserve Pointer references if peer unchanged", p, s.p);
+    }
 
     public void testOverwriteStructureByReferenceFieldOnRead() {
         StructureWithPointers s = new StructureWithPointers();
@@ -756,14 +830,6 @@ public class StructureTest extends TestCase {
     	s.write();
     }
 
-    public static class TestNativeMappedInStructure extends Structure {
-        public static class ByValue extends TestNativeMappedInStructure implements Structure.ByValue { }
-        public NativeLong field;
-    }
-    public void testNativeMappedInByValue() {
-        new TestNativeMappedInStructure.ByValue();
-    }
-
     public static class ROStructure extends Structure {
         public final int field;
         {
@@ -802,4 +868,117 @@ public class StructureTest extends TestCase {
                          i, s.longs[i].intValue());
         }
     }
+
+
+    public void testInitializeNativeMappedField() {
+        final long VALUE = 20;
+        final NativeLong INITIAL = new NativeLong(VALUE);
+        class TestStructure extends Structure {
+            // field overwritten, wrong value before write
+            // NL bug, wrong value written
+            { setAlignType(ALIGN_NONE); }
+            public NativeLong nl = INITIAL;
+            public NativeLong uninitialized;
+        }
+        TestStructure ts = new TestStructure();
+        assertEquals("Wrong value in field", VALUE, ts.nl.longValue());
+        assertSame("Initial value overwritten", INITIAL, ts.nl);
+        assertEquals("Wrong field value before write", VALUE, ts.nl.longValue());
+        assertNotNull("Uninitialized field should be initialized", ts.uninitialized);
+        assertEquals("Wrong initialized value", 0, ts.uninitialized.longValue());
+        ts.write();
+        assertEquals("Wrong field value written", VALUE, ts.getPointer().getNativeLong(0).longValue());
+        assertEquals("Wrong field value written (2)", 0, ts.getPointer().getNativeLong(NativeLong.SIZE).longValue());
+        ts.read();
+        assertEquals("Wrong field value", VALUE, ts.nl.longValue());
+        assertEquals("Wrong field value (2)", 0, ts.uninitialized.longValue());
+    }
+
+    public void testStructureFieldOrder() {
+        Structure.REQUIRES_FIELD_ORDER = true;
+        try {
+            class TestStructure extends Structure {
+                public int one = 1;
+                public int three = 3;
+                public int two = 2;
+                {
+                    setFieldOrder(new String[] { "one", "two", "three" });
+                }
+            }
+            class DerivedTestStructure extends TestStructure {
+                public int four = 4;
+                {
+                    setFieldOrder(new String[] { "four" });
+                }
+            }
+
+            DerivedTestStructure s = new DerivedTestStructure();
+            DerivedTestStructure s2 = new DerivedTestStructure();
+            s.write();
+            s2.write();
+            assertEquals("Wrong first field", 1, s.getPointer().getInt(0));
+            assertEquals("Wrong second field", 2, s.getPointer().getInt(4));
+            assertEquals("Wrong third field", 3, s.getPointer().getInt(8));
+            assertEquals("Wrong derived field", 4, s.getPointer().getInt(12));
+        }
+        finally {
+            Structure.REQUIRES_FIELD_ORDER = false;
+        }
+    }
+
+    public void testCustomTypeMapper() {
+        class TestField { }
+        class TestStructure extends Structure {
+            public TestField field;
+            public TestStructure() {
+                DefaultTypeMapper m = new DefaultTypeMapper();
+                m.addTypeConverter(TestField.class, new TypeConverter() {
+                    public Object fromNative(Object value, FromNativeContext context) {
+                        return new TestField();
+                    }
+                    public Class nativeType() {
+                        return String.class;
+                    }
+                    public Object toNative(Object value, ToNativeContext ctx) {
+                        return value == null ? null : value.toString();
+                    }
+                });
+                setTypeMapper(new DefaultTypeMapper());
+            }
+        }
+        new TestStructure();
+    }
+    
+    public void testWriteWithNullBoxedPrimitives() {
+        class TestStructure extends Structure {
+            public Boolean zfield;
+            public Integer field;
+        }
+        TestStructure s = new TestStructure();
+        s.write();
+        s.read();
+        assertNotNull("Field should not be null after read", s.field);
+    }
+
+    public interface AutoSynchTest extends Library {
+        public static class TestStructure extends Structure {
+            public int field;
+        }
+        Pointer testStructurePointerArgument(TestStructure s);
+    }
+    public void testDisableAutoSynch() {
+        AutoSynchTest lib = (AutoSynchTest)Native.loadLibrary("testlib", AutoSynchTest.class);
+        AutoSynchTest.TestStructure s = new AutoSynchTest.TestStructure();
+        final int VALUE = 42;
+        s.field = VALUE;
+        s.setAutoWrite(false);
+        lib.testStructurePointerArgument(s);
+        assertEquals("Auto write should be disabled", 0, s.field);
+
+        final int EXPECTED = s.field;
+        s.getPointer().setInt(0, VALUE);
+        s.setAutoRead(false);
+        lib.testStructurePointerArgument(s);
+        assertEquals("Auto read should be disabled", EXPECTED, s.field);
+    }
 }
\ No newline at end of file
